#!/bin/bash

# Global variables
ns3_ver=3.16
netanim_ver=3.103
py="/usr/bin/python2"

build_targets=("netanim" "ns3" "all")
clean_targets=${build_targets[*]}
ns3_targets=("debug" "release" "optimized")

ns3_modules="diffserv"

backup_targets=("repo" "full")
repo_list="file.list"

# Functions
function usage()
{
  echo "usage: $0 action [argument(s)]"
  echo "------------------"
  echo "available actions:"
  echo "------------------"
  echo "# targets: (${build_targets[*]})"
  echo "# ns3_targets: (${ns3_targets[*]})"
  echo "build target  [ns3_target [python]]"
  echo "clean target (${clean_targets[*]})"
  echo "run file"
  echo "pyrun file"
  echo "gdb_run file"
  echo "val_run file"
  echo "dump file"
}

function error_msg()
{
  echo "invalid ${msg}: ${msg_error}"
  echo ""
  usage
  exit
}

function no_msg()
{
  echo "no ${msg} specified."
  echo ""
  usage
  exit
}

# Build
function configure_ns3()
{
  echo "###################"
  echo "# Configuring: ns-3"
  echo "###################"
  cd src/ns-allinone-${ns3_ver}/ns-${ns3_ver}
  ${py} waf configure -d ${ns3_target} -o build/${ns3_target} --with-python=${py}
  cd ../../..
}

function build_ns3()
{
  echo "################"
  echo "# Building: ns-3"
  echo "################"
  cd src/ns-allinone-${ns3_ver}/ns-${ns3_ver}
  ${py} waf
  cd ../../..
}

function build_bindings()
{
  echo "#########################"
  echo "# Building: ns-3 bindings"
  echo "#########################"
  cd src/ns-allinone-${ns3_ver}/ns-${ns3_ver}
  ${py} waf --apiscan=${ns3_modules}
  cd ../../..
}

function build_netanim()
{
  echo "###################"
  echo "# Building: NetAnim"
  echo "###################"
  cd src/ns-allinone-${ns3_ver}/netanim-${netanim_ver}
  qmake-qt4 NetAnim.pro
  make
  cd ../../..
}

# Clean
function clean_ns3()
{
  echo "###############"
  echo "# Cleaning ns-3"
  echo "###############"
  cd src/ns-allinone-${ns3_ver}/ns-${ns3_ver}
  ${py} waf clean
  cd ../../..
}

function clean_netanim()
{
  echo "##################"
  echo "# Cleaning NetAnim"
  echo "##################"
  cd src/ns-allinone-${ns3_ver}/netanim-${netanim_ver}
  make clean
  cd ../../..
}

# Run
function run()
{
  cd src/ns-allinone-${ns3_ver}/ns-${ns3_ver}
  ${py} waf --run "${file}" --cwd="../../../traces"
  cd ../../..
}

function pyrun()
{
  cd src/ns-allinone-${ns3_ver}/ns-${ns3_ver}
  ${py} waf --pyrun "${file}" --cwd="../../../traces"
  cd ../../..
}

function gdb_run()
{
  cd src/ns-allinone-${ns3_ver}/ns-${ns3_ver}
  ${py} waf --run "${file}" --cwd="../../../traces" \
  --command-template="gdb --args %s"
  cd ../../..
}

function val_run()
{
  cd src/ns-allinone-${ns3_ver}/ns-${ns3_ver}
  ${py} waf --run "${file}" --cwd="../../../traces" \
  --command-template="valgrind --track-origins=yes %s"
  # --leak-check=full --show-reachable=yes
  cd ../../..
}

# Dump
function dump()
{
  tcpdump -nn -tt -r ${file}
}

# main
if [[ $# -lt 1 ]]
then
  msg="action"
  no_msg
  exit
fi

action=$1
case ${action} in
build)
  if [[ $# -lt 2 ]]
  then
    msg="target"
    no_msg
  fi

  target=$2
  case ${target} in
  ns3)
    if [[ $# -lt 3 ]]
    then
      msg="ns3_target"
      no_msg
    fi

    ns3_target=$3
    flag=0
    for i in "${ns3_targets[@]}"
    do
      if [[ $i == ${ns3_target} ]]
      then
        flag=1
      fi
    done

    py_flag=0
    if [[ $# -eq 4 ]]
    then
      if [[ $4 -eq "python" ]]
      then
        py_flag=1
      fi
    fi

    if [[ ${flag} -eq 1 ]]
    then
      configure_ns3
      if [[ ${py_flag} -eq 1 ]]
      then
        build_bindings
      fi
      build_ns3
    else
      msg="ns3_target"
      msg_error=${ns3_target}
      error_msg
    fi
  ;;
  netanim)
    build_netanim
  ;;
  all)
    if [[ $# -lt 3 ]]
    then
      msg="ns3_target"
      no_msg
    fi

    ns3_target=$3
    flag=0
    for i in "${ns3_targets[@]}"
    do
      if [[ $i == ${ns3_target} ]]
      then
        flag=1
      fi
    done

    py_flag=0
    if [[ $# -eq 4 ]]
    then
      if [[ $4 -eq "python" ]]
      then
        py_flag=1
      fi
    fi

    if [[ ${flag} -eq 1 ]]
    then
      build_netanim
      configure_ns3
      if [[ ${py_flag} -eq 1 ]]
      then
        build_bindings
      fi
      build_ns3
    else
      msg="ns3_target"
      msg_error=${ns3_target}
      error_msg
    fi
  ;;
  *)
    msg="target"
    msg_error=${target}
    error_msg
  ;;
  esac
;;
clean)
  if [[ $# -lt 2 ]]
  then
    msg="target"
    no_msg
  fi

  target=$2
  case ${target} in
  ns3)
    clean_ns3
  ;;
  netanim)
    clean_netanim
  ;;
  all)
    clean_ns3
    clean_netanim
  ;;
  *)
    msg="target"
    msg_error=${target}
    error_msg
  ;;
  esac
;;
run)
  if [[ $# -lt 2 ]]
  then
    msg="file"
    no_msg
  fi

  file=$2
  run
;;
pyrun)
  if [[ $# -lt 2 ]]
  then
    msg="file"
    no_msg
  fi

  file=$2
  pyrun
;;
gdb_run)
  if [[ $# -lt 2 ]]
  then
    msg="file"
    no_msg
  fi

  file=$2
  gdb_run
;;
val_run)
  if [[ $# -lt 2 ]]
  then
    msg="file"
    no_msg
  fi

  file=$2
  val_run
;;
dump)
  if [[ $# -lt 2 ]]
  then
    msg="file"
    no_msg
  fi

  file=$2
  dump
;;
*)
  msg="action"
  msg_error=${action}
  error_msg
;;
esac
