#!/usr/bin/env python2
# -*- coding: utf-8 -*-

import xml.etree.ElementTree as ElementTree
import argparse
import sys
import os

NetProtocol = {6: 'TCP', 17: 'UDP'}
LatexFile = "Tests.tex"

###########
# Metrics #
###########
# tx_duration = (timeLastRxPacket - timeFirstTxPacket) (sec)
def tx_duration(timeFirstTxPacket, timeLastRxPacket):
    if ns2ns(timeLastRxPacket) == 0:
        return 0
    else:
        return round((ns2ns(timeLastRxPacket)-ns2ns(timeFirstTxPacket)) * 1e-9, 3)

# mean_delay = (delaySum / rxPackets) (ms)
def mean_delay(delaySum, rxPackets):
    if ns2ns(rxPackets) == 0:
        return 0
    else:
        return round((ns2ns(delaySum) / ns2ns(rxPackets)) * 1e-6, 3)

# mean_jitter = (jitterSum / (rxPackets - 1)) (μsec)
def mean_jitter(jitterSum, rxPackets):
    if ns2ns(rxPackets) == 0:
        return 0
    else:
        return round((ns2ns(jitterSum) / (ns2ns(rxPackets) - 1)) * 1e-3, 3)

# mean_tx_packet_size = txBytes / txPackets (bytes)
def mean_tx_packet_size(txBytes, txPackets):
    return round(ns2ns(txBytes) / ns2ns(txPackets), 3)

# mean_rx_packet_size = rxBytes / rxPackets (bytes)
def mean_rx_packet_size(rxBytes, rxPackets):
    if ns2ns(rxPackets) == 0:
        return 0
    else:
        return round(ns2ns(rxBytes) / ns2ns(rxPackets), 3)

# mean_tx_bitrate = [(8 * txBytes) / (timeLastTxPacket - timeFirstTxPacket)] (Kbps)
def mean_tx_bitrate(txBytes, timeFirstTxPacket, timeLastTxPacket):
    bitrate = (8 * long(txBytes)) / ((ns2ns(timeLastTxPacket)-ns2ns(timeFirstTxPacket)) * 1e-9)
    return round(bitrate/1024, 3)

# mean_rx_bitrate = [(8 * rxBytes) / (timeLastRxPacket - timeFirstRxPacket)] (Kbps)
def mean_rx_bitrate(rxBytes, timeFirstRxPacket, timeLastRxPacket):
    if ns2ns(rxBytes) == 0:
        return 0
    else:
        bitrate = (8 * long(rxBytes)) / ((ns2ns(timeLastRxPacket)-ns2ns(timeFirstRxPacket)) * 1e-9)
        return round(bitrate/1024, 3)

# mean_hop_count = 1 + (timesForwarded / rxPackets)
def mean_hop_count(timesForwarded, rxPackets):
    if ns2ns(rxPackets) == 0:
        return 0
    else:
        return round(1 + (ns2ns(timesForwarded) / ns2ns(rxPackets)), 3)

# throughput = [(rxBytes*8) / (LastRxPacket-FirstTxPacket)] (Kbps)
def throughput(rxBytes, timeFirstTxPacket, timeLastRxPacket):
    if ns2ns(rxBytes) == 0:
        return 0
    else:
        duration = (ns2ns(timeLastRxPacket) - ns2ns(timeFirstTxPacket)) * 1e-9
        throughput = ((long(rxBytes) * 8) / duration)
        return round(throughput/1024, 3)

# Packet delivery ratio
# pdr = rxPackets / txPackets
def pdr(txPackets, rxPackets):
    return round(float(rxPackets)/float(txPackets), 3)

# Packet loss ratio
# plr = lostPackets / txPackets
def plr(txPackets, lostPackets):
    return round(float(lostPackets)/float(txPackets), 3)

####################
# Helper functions #
####################
# Convert nanoseconds string to float
def ns2ns(ns):
    return float(ns.split('ns')[0])

# Convert nanoseconds string to miliseconds float
def ns2ms(ns):
    return round(ns2ns(ns) * 1e-6, 3)

# Parse flowmitor's XML
def parse_xml(input_file):
    Results = {}

    print ""
    print "parsing:", input_file
    root = ElementTree.parse(input_file)
    print "done!"
    print ""

    Results['Ids'] = []
    FlowStats = root.find('FlowStats')
    Flow = FlowStats.iterfind('Flow')
    for x in Flow:
        flowId = x.get('flowId')
        timeFirstTxPacket = x.get('timeFirstTxPacket')
        timeFirstRxPacket = x.get('timeFirstRxPacket')
        timeLastTxPacket = x.get('timeLastTxPacket')
        timeLastRxPacket = x.get('timeLastRxPacket')
        delaySum = x.get('delaySum')
        jitterSum = x.get('jitterSum')
        lastDelay = x.get('lastDelay')
        txBytes = x.get('txBytes')
        rxBytes = x.get('rxBytes')
        txPackets = x.get('txPackets')
        rxPackets = x.get('rxPackets')
        lostPackets = x.get('lostPackets')
        timesForwarded = x.get('timesForwarded')

        Results['Ids'].append(flowId)
        Results[flowId] = {}
        Results[flowId]['FlowStats'] = {}
        Results[flowId]['FlowStats']['timeFirstTxPacket'] = timeFirstTxPacket
        Results[flowId]['FlowStats']['timeFirstRxPacket'] = timeFirstRxPacket
        Results[flowId]['FlowStats']['timeLastTxPacket'] = timeLastTxPacket
        Results[flowId]['FlowStats']['timeLastRxPacket'] = timeLastRxPacket
        Results[flowId]['FlowStats']['delaySum'] = delaySum
        Results[flowId]['FlowStats']['jitterSum'] = jitterSum
        Results[flowId]['FlowStats']['lastDelay'] = lastDelay
        Results[flowId]['FlowStats']['txBytes'] = txBytes
        Results[flowId]['FlowStats']['rxBytes'] = rxBytes
        Results[flowId]['FlowStats']['txPackets'] = txPackets
        Results[flowId]['FlowStats']['rxPackets'] = rxPackets
        Results[flowId]['FlowStats']['lostPackets'] = lostPackets
        Results[flowId]['FlowStats']['timesForwarded'] = timesForwarded

        PacketsDropped = {}
        packetsDropped = x.iterfind('packetsDropped')
        for i in packetsDropped:
            reasonCode = i.get('reasonCode')
            number = i.get('number')
            PacketsDropped[reasonCode] = number
        Results[flowId]['PacketsDropped'] = PacketsDropped

        BytesDropped = {}
        bytesDropped = x.iterfind('bytesDropped')
        for i in bytesDropped:
            reasonCode = i.get('reasonCode')
            number = i.get('bytes')
            BytesDropped[reasonCode] = number
        Results[flowId]['BytesDropped'] = BytesDropped

        DelayInfo = []
        delayHistogram = x.find('delayHistogram')
        nBins = delayHistogram.get('nBins')
        Results[flowId]['DelayBins'] = nBins

        bin = delayHistogram.iterfind('bin')
        for i in bin:
            start = ns2ns(i.get('start'))
            width = ns2ns(i.get('width'))
            count = ns2ns(i.get('count'))
            DelayInfo.append([start, width, count])
        Results[flowId]['DelayInfo'] = DelayInfo

        JitterInfo = []
        jitterHistogram = x.find('jitterHistogram')
        nBins = jitterHistogram.get('nBins')
        Results[flowId]['JitterBins'] = nBins

        bin = jitterHistogram.iterfind('bin')
        for i in bin:
            start = ns2ns(i.get('start'))
            width = ns2ns(i.get('width'))
            count = ns2ns(i.get('count'))
            JitterInfo.append([start, width, count])
        Results[flowId]['JitterInfo'] = JitterInfo

        PacketSizeInfo = []
        packetSizeHistogram = x.find('packetSizeHistogram')
        nBins = packetSizeHistogram.get('nBins')
        Results[flowId]['PacketSizeBins'] = nBins

        bin = packetSizeHistogram.iterfind('bin')
        for i in bin:
            start = ns2ns(i.get('start'))
            width = ns2ns(i.get('width'))
            count = ns2ns(i.get('count'))
            PacketSizeInfo.append([start, width, count])
        Results[flowId]['PacketSizeInfo'] = PacketSizeInfo

        FlowInterruptions = []
        flowInterruptionsHistogram = x.find('flowInterruptionsHistogram')
        nBins = flowInterruptionsHistogram.get('nBins')
        Results[flowId]['FlowInterruptionsBins'] = nBins

        bin = flowInterruptionsHistogram.iterfind('bin')
        for i in bin:
            start = ns2ns(i.get('start'))
            width = ns2ns(i.get('width'))
            count = ns2ns(i.get('count'))
            FlowInterruptions.append([start, width, count])
        Results[flowId]['FlowInterruptionsInfo'] = FlowInterruptions

    Ipv4FlowClassifier = root.find('Ipv4FlowClassifier')
    Flow = Ipv4FlowClassifier.iterfind('Flow')
    for x in Flow:
        flowId = x.get('flowId')
        sourceAddress = x.get('sourceAddress')
        destinationAddress = x.get('destinationAddress')
        protocol = x.get('protocol')
        sourcePort = x.get('sourcePort')
        destinationPort = x.get('destinationPort')

        Results[flowId]['Ipv4FlowClassifier'] = {}
        Results[flowId]['Ipv4FlowClassifier']['SrcAddress'] = sourceAddress
        Results[flowId]['Ipv4FlowClassifier']['DestAddress'] = destinationAddress
        Results[flowId]['Ipv4FlowClassifier']['Protocol'] = protocol
        Results[flowId]['Ipv4FlowClassifier']['SrcPort'] = sourcePort
        Results[flowId]['Ipv4FlowClassifier']['DestPort'] = destinationPort

    Results['Probes'] = {}
    FlowProbes = root.find('FlowProbes')
    FlowProbe = FlowProbes.iterfind('FlowProbe')
    for x in FlowProbe:
        index = x.get('index')
        Results['Probes'][index] = {}

        FlowStats = x.iterfind('FlowStats')
        for y in FlowStats:
            flowId = y.get('flowId')
            packets = y.get('packets')
            bytes = y.get('bytes')
            delayFromFirstProbeSum = y.get('delayFromFirstProbeSum')

            Probes = {}
            Probes['Packets'] = packets
            Probes['Bytes'] = bytes
            Probes['DelaySum'] = delayFromFirstProbeSum

            packetsDropped = y.iterfind('packetsDropped')
            packets = {}
            for z in packetsDropped:
                reasonCode = z.get('reasonCode')
                number = z.get('number')
                packets[reasonCode] = number
            Probes['PacketsDropped'] = packets

            bytesDropped = y.iterfind('bytesDropped')
            bytes = {}
            for z in bytesDropped:
                reasonCode = z.get('reasonCode')
                number = z.get('bytes')
                bytes[reasonCode] = number
            Probes['BytesDropped'] = bytes

            Results['Probes'][index][flowId] = Probes

    for fid in sorted(Results['Ids']):
        i = Results[fid]['FlowStats']
        TxDuration = tx_duration(i['timeFirstTxPacket'], i['timeLastRxPacket'])
        MeanHopCount = mean_hop_count(i['timesForwarded'], i['rxPackets'])
        Throughput = throughput(i['rxBytes'], i['timeFirstTxPacket'], i['timeLastRxPacket'])
        MeanTxBitrate = mean_tx_bitrate(i['txBytes'], i['timeFirstTxPacket'], i['timeLastTxPacket'])
        MeanRxBitrate = mean_rx_bitrate(i['rxBytes'], i['timeFirstRxPacket'], i['timeLastRxPacket'])
        MeanDelay =mean_delay(i['delaySum'], i['rxPackets'])
        MeanJitter = mean_jitter(i['jitterSum'], i['rxPackets'])
        MeanTxPacketSize = mean_tx_packet_size(i['txBytes'], i['txPackets'])
        MeanRxPacketSize = mean_rx_packet_size(i['rxBytes'], i['rxPackets'])
        Pdr = pdr(i['txPackets'], i['rxPackets'])
        Plr = plr(i['txPackets'], i['lostPackets'])

        Results[fid]['Metrics'] = {}
        Results[fid]['Metrics']['TxDuration'] = TxDuration
        Results[fid]['Metrics']['MeanHopCount'] = MeanHopCount
        Results[fid]['Metrics']['Throughput'] = Throughput
        Results[fid]['Metrics']['MeanTxBitrate'] = MeanTxBitrate
        Results[fid]['Metrics']['MeanRxBitrate'] = MeanRxBitrate
        Results[fid]['Metrics']['MeanDelay'] = MeanDelay
        Results[fid]['Metrics']['MeanJitter'] = MeanJitter
        Results[fid]['Metrics']['MeanTxPacketSize'] = MeanTxPacketSize
        Results[fid]['Metrics']['MeanRxPacketSize'] = MeanRxPacketSize
        Results[fid]['Metrics']['Pdr'] = Pdr
        Results[fid]['Metrics']['Plr'] = Plr

    return Results

# Print parsed data
def print_xml(Results, verbose):
    print "#############################"
    print ">>> Simulation statistics <<<"
    print "#############################"
    for fid in Results['Ids']:
        print "+++++++++++++"
        print "+ FlowID:", fid, "+"
        print "+++++++++++++"
        i = Results[fid]['Ipv4FlowClassifier']
        print "---------------------"
        print "Flow Ipv4 info"
        print "---------------------"
        print "SrcAddress :", i['SrcAddress']
        print "DestAddress:", i['DestAddress']
        print "Protocol   :", i['Protocol'], "(" + NetProtocol[int(i['Protocol'])] + ")"
        print "SrcPort    :", i['SrcPort']
        print "DestPort   :", i['DestPort']
        print ""

        i = Results[fid]['FlowStats']
        print "---------------------------------"
        print "Flow statistics"
        print "---------------------------------"
        print "timeFirstTxPacket :", ns2ms(i['timeFirstTxPacket']), "ms"
        print "timeFirstRxPacket :", ns2ms(i['timeFirstRxPacket']), "ms"
        print "timeLastTxPacket  :", ns2ms(i['timeLastTxPacket']), "ms"
        print "timeLastRxPacket  :", ns2ms(i['timeLastRxPacket']), "ms"
        print "delaySum          :", ns2ms(i['delaySum']), "ms"
        print "jitterSum         :", ns2ms(i['jitterSum']), "ms"
        print "txBytes           :", i['txBytes'], "bytes"
        print "rxBytes           :", i['rxBytes'], "bytes"
        print "txPackets         :", i['txPackets']
        print "rxPackets         :", i['rxPackets']
        print "lostPackets       :", i['lostPackets']
        print "timesForwarded    :", i['timesForwarded']
        print ""

        if verbose >= 1:
            i = Results[fid]['PacketsDropped']
            ik = sorted(i.keys())
            j = Results[fid]['BytesDropped']
            print "---------------------"
            print "Dropped packets"
            print "---------------------"
            print "Code\tNumber\tBytes"
            print "---------------------"
            if not ik:
                print "-- None --"
            else:
                for c in ik:
                    print c, "\t", i[c], "\t", j[c]
            print ""

            i = Results['Probes']
            ik = sorted(i.keys())
            print "-------------------------------"
            print "Dropped packets probes"
            print "-------------------------------"
            print "Probe\tCode\tNumber\tBytes"
            print "-------------------------------"
            ctr = 0
            for c in ik:
                if fid in i[c].keys():
                    iik = sorted(i[c][fid]['PacketsDropped'].keys())
                    for cc in iik:
                        print c, cc, i[c][fid]['PacketsDropped'][cc], i[c][fid]['BytesDropped'][cc]
                        ctr += 1
            if ctr == 0:
                print "-- None --"
            print ""

            print "-------------------------------------------------------------"
            print "Flow probes info"
            print "-------------------------------------------------------------"
            print "Probe\tPackets\t\tBytes\t\tDelaySum (ms)"
            print "-------------------------------------------------------------"
            ctr = 0
            for c in ik:
                if fid in i[c].keys():
                    print c, "\t", i[c][fid]['Packets'], "\t\t", i[c][fid]['Bytes'], \
                            "\t", ns2ms(i[c][fid]['DelaySum']), "ms"
                    ctr += 1
            if ctr == 0:
                print "-- None --"
            print ""

        if verbose >= 2:
            i = Results[fid]['DelayInfo']
            print "---------------------"
            print "Delay histogram"
            print "---------------------"
            print "Start\tWidth\tCount"
            print "---------------------"
            if not i:
                print "-- None --"
            else:
                for c in i:
                    print c[0], "\t", c[1], "\t", c[2]
            print ""

            i = Results[fid]['JitterInfo']
            print "---------------------"
            print "Jitter histogram"
            print "---------------------"
            print "Start\tWidth\tCount"
            print "---------------------"
            if not i:
                print "-- None --"
            else:
                for c in i:
                    print c[0], "\t", c[1], "\t", c[2]
            print ""

            i = Results[fid]['PacketSizeInfo']
            print "---------------------"
            print "Packet size histogram"
            print "---------------------"
            print "Start\tWidth\tCount"
            print "---------------------"
            if not i:
                print "-- None --"
            else:
                for c in i:
                    print c[0], "\t", c[1], "\t", c[2]
            print ""

            i = Results[fid]['FlowInterruptionsInfo']
            print "----------------------------"
            print "Flow interruptions histogram"
            print "----------------------------"
            print "Start\tWidth\tCount"
            print "----------------------------"
            if not i:
                print "-- None --"
            else:
                for c in i:
                    print c[0], "\t", c[1], "\t", c[2]
            print ""

        i = Results[fid]['FlowStats']
        TxDuration = tx_duration(i['timeFirstTxPacket'], i['timeLastRxPacket'])
        MeanHopCount = mean_hop_count(i['timesForwarded'], i['rxPackets'])
        Throughput = throughput(i['rxBytes'], i['timeFirstTxPacket'], i['timeLastRxPacket'])
        MeanTxBitrate = mean_tx_bitrate(i['txBytes'], i['timeFirstTxPacket'], i['timeLastTxPacket'])
        MeanRxBitrate = mean_rx_bitrate(i['rxBytes'], i['timeFirstRxPacket'], i['timeLastRxPacket'])
        MeanDelay =mean_delay(i['delaySum'], i['rxPackets'])
        MeanJitter = mean_jitter(i['jitterSum'], i['rxPackets'])
        MeanTxPacketSize = mean_tx_packet_size(i['txBytes'], i['txPackets'])
        MeanRxPacketSize = mean_rx_packet_size(i['rxBytes'], i['rxPackets'])
        Pdr = pdr(i['txPackets'], i['rxPackets'])
        Plr = plr(i['txPackets'], i['lostPackets'])

        print "------------------------------------"
        print "Flow metrics"
        print "------------------------------------"
        print "Tx duration          :", Results[fid]['Metrics']['TxDuration'], "sec"
        print "Mean hop count       :", Results[fid]['Metrics']['MeanHopCount']
        print "------------------------------------"
        print "Throughput           :", Results[fid]['Metrics']['Throughput'], "Kbps"
        print "Mean Tx bitrate      :", Results[fid]['Metrics']['MeanTxBitrate'], "Kbps"
        print "Mean Rx bitrate      :", Results[fid]['Metrics']['MeanRxBitrate'], "Kbps"
        print "Mean delay           :", Results[fid]['Metrics']['MeanDelay'], "ms"
        print "Mean jitter          :", Results[fid]['Metrics']['MeanJitter'], "μsec"
        print "------------------------------------"
        print "Mean Tx packet size  :", Results[fid]['Metrics']['MeanTxPacketSize'], "bytes"
        print "Mean Rx packet size  :", Results[fid]['Metrics']['MeanRxPacketSize'], "bytes"
        print "Packet Delivery Ratio:", Results[fid]['Metrics']['Pdr']
        print "Packet Loss Ratio    :", Results[fid]['Metrics']['Plr']
        print "------------------------------------"
        print ""

    if not Results.keys():
        print "-- None --"
    print "####################################"
    print ">>> End of Simulation statistics <<<"
    print "####################################"
    print ""

# Convert XML traces to LaTeX tables
def convert_xml(Results):
    begin_flushleft = "\\begin{flushleft}"+"\n"
    end_flushleft = "\end{flushleft}"+"\n"
    begin_table = "\\begin{table}[t]\\footnotesize"+"\n"
    end_table = "\end{table}"+"\n"
    begin_minipage = "\\begin{minipage}{0.3\linewidth}"+"\n"
    end_minipage = "\end{minipage}"+"\n"
    quad = "\quad"+"\n"

    page = ""
    table = begin_flushleft+begin_table
    ctr = 1
    for i in Results:
        file = i[1]
        TestName = os.path.splitext(file)[0]
        for fid in i[0]['Ids']:
            info = i[0][fid]['Ipv4FlowClassifier']
            metrics = i[0][fid]['Metrics']
            minipage = begin_minipage
            tabular = "\\begin{tabular}{|l|l|}"+"\n"
            tabular += "\hline"+"\n"
            tabular += "\multicolumn{2}{|c|}{FlowID: "+ fid +"} \\\ \hline\hline"+"\n"
            tabular += "SrcAddress & "+ info['SrcAddress'] +" \\\ \hline"+"\n"
            tabular += "DestAddress & "+ info['DestAddress'] +" \\\ \hline"+"\n"
            tabular += "Protocol & "+ info['Protocol'] +" ("+ NetProtocol[int(info['Protocol'])] +")"+" \\\ \hline"+"\n"
            tabular += "SrcPort & "+ info['SrcPort'] +" \\\ \hline"+"\n"
            tabular += "DestPort & "+ info['DestPort'] +" \\\ \hline"+"\n"
            tabular += "TxDuration & "+ str(metrics['TxDuration']) +" sec \\\ \hline"+"\n"
            tabular += "Throughput & "+ str(metrics['Throughput']) +" Kbps \\\ \hline"+"\n"
            tabular += "\(\overline{Tx bitrate}\) & "+ str(metrics['MeanTxBitrate']) +" Kbps \\\ \hline"+"\n"
            tabular += "\(\overline{Rx bitrate}\) & "+ str(metrics['MeanRxBitrate']) +" Kbps \\\ \hline"+"\n"
            tabular += "\(\overline{delay}\) & "+ str(metrics['MeanDelay']) +" ms \\\ \hline"+"\n"
            tabular += "\(\overline{jitter}\) & "+ str(metrics['MeanJitter']) +" μsec \\\ \hline"+"\n"
            tabular += "\(\overline{Tx packet size}\) & "+ str(metrics['MeanTxPacketSize']) +" bytes \\\ \hline"+"\n"
            tabular += "\(\overline{Rx packet size}\) & "+ str(metrics['MeanRxPacketSize']) +" bytes \\\ \hline"+"\n"
            tabular += "PDR & "+ str(metrics['Pdr']) +" \\\ \hline"+"\n"
            tabular += "PLR & "+ str(metrics['Plr']) +" \\\ \hline"+"\n"
            tabular += "\end{tabular}"+"\n"

            caption = "\caption{"+TestName+", Flow:"+fid+".}"+"\n"
            label = "\label{tab:"+TestName+"-Flow"+fid+"}"+"\n"
            tabular += caption+label
            minipage += tabular+end_minipage

            if ctr < 6:
                minipage += quad
                table += minipage
                ctr += 1
            else:
                minipage += end_table
                table += minipage
                page += table
                table = begin_table
                ctr = 1

    if ctr > 1:
        table += end_table
        page += table
    page += end_flushleft
    with open(LatexFile, 'w') as fp:
        fp.write(page)

########
# main #
########
def main():
    parser = argparse.ArgumentParser(description="ns-3 flowmonitor xml parser",
                                     epilog="Happy parsing! :P")
    parser.add_argument("-i", "--input", help="Parse: XML trace file.")
    parser.add_argument("-l", "--latex",
                        help="Convert directory's XML traces to LaTeX.")
    parser.add_argument("-v", "--verbose", type=int, default=0,
                        help="Turn verbosity: [0-2](default: 0).")

    args = parser.parse_args()

    verbose = args.verbose
    if verbose < 0 or verbose > 2:
        print "Out of bounds verbose value!"
        print ""
        parser.print_help()
        sys.exit()

    input_file = args.input
    latex_dir = args.latex

    if not input_file and not latex_dir:
        print "No action specified."
        print "Try: -i|--input OR -l|--latex."
        print ""
        parser.print_help()
        sys.exit()

    if input_file:
        if not os.path.isfile(input_file):
            print "Argument:", input_file, "is not a file."
            print ""
            parser.print_help()
            sys.exit()

        Results = parse_xml(input_file)
        print_xml(Results, verbose)

    if latex_dir:
        if not os.path.isdir(latex_dir):
            print "Argument:", input_file, "is not a directory."
            print ""
            parser.print_help()
            sys.exit()

        Results = []
        files = os.listdir(latex_dir)
        for i in sorted(files):
            if os.path.isfile(i) and os.path.splitext(i)[1] == '.xml':
                Results.append([parse_xml(i), i])
        convert_xml(Results)

if __name__ == "__main__":
    sys.exit(main())