/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2013 University of Patras,
 * Department of Computer Engineering & Informatics
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * Author: Christos Nianiakas <cnianiak_at_gmail_dot_com>
 */

#include "ns3/diffserv-module.h"
#include "ns3/core-module.h"
#include "ns3/network-module.h"
#include "ns3/internet-module.h"
#include "ns3/point-to-point-module.h"
#include "ns3/applications-module.h"
#include "ns3/mobility-module.h"
#include "ns3/flow-monitor-module.h"
#include "ns3/netanim-module.h"
#include <vector>

// Network Topology
//     10.1.1.0                                                 10.3.1.0
// n0 ----------+                                             +---------- n3
//     10.1.2.0 |           10.2.1.0       10.2.2.0           | 10.3.2.0
// n1 ----------+----- ds0 ---------- ds1 ---------- ds2 -----+---------- n4
//     10.1.3.0 | p2p         p2p            p2p          p2p | 10.3.3.0
// n2 ----------+                                             +---------- n5
//
// p2p: point-to-point
// n0 VoIP traffic to n3
// n1 VoIP, Videotrace traffic to n4
// n2 VoIP traffic to n5
// ds0: DiffServ domain (ingress node)
// ds1: DiffServ domain (core node)
// ds2: DiffServ domain (egress node)
// n3: Packet sink (for n0)
// n4: Packet sink (for n1)
// n5: Packet sink (for n3)

using namespace ns3;

#define ScriptName "DiffServSimulator"

NS_LOG_COMPONENT_DEFINE (ScriptName);

void PrintSimTime (void);

int
main (int argc, char *argv[])
{
  // Simulation variables
  std::string SimName = "Test";

  double SimDuration = 60.0;
  double SimStart = 0;

  uint32_t Verbose = 0;
  uint32_t Pcap = 0;
  uint32_t Ascii = 0;
  uint32_t Checksum = 0;

  uint32_t NetAnim = 0;
  uint32_t Metadata = 0;
  uint32_t Flowmonitor = 1;

  uint32_t DS = 0;
  uint32_t Policy = 0;
  std::string MdrrMode = "ALTERNATE";
  uint16_t MTU = 1500;
  uint32_t EFsize = 100;
  uint32_t AF1size = 100;
  uint32_t AF2size = 100;
  uint32_t AF3size = 100;
  uint32_t AF4size = 100;
  uint32_t BEsize = 100;
  uint32_t EFweight = 1;
  uint32_t AF1weight = 1;
  uint32_t AF2weight = 1;
  uint32_t AF3weight = 1;
  uint32_t AF4weight = 1;
  uint32_t BEweight = 1;

  DataRate n0ds0Rate ("2Mbps");
  DataRate n1ds0Rate ("2Mbps");
  DataRate n2ds0Rate ("2Mbps");
  DataRate ds0ds1Rate ("20Mbps");
  DataRate ds1ds2Rate ("20Mbps");
  DataRate ds2n3Rate ("2Mbps");
  DataRate ds2n4Rate ("2Mbps");
  DataRate ds2n5Rate ("2Mbps");

  Time n0ds0Delay ("10ms");
  Time n1ds0Delay ("10ms");
  Time n2ds0Delay ("10ms");
  Time ds0ds1Delay ("20ms");
  Time ds1ds2Delay ("20ms");
  Time ds2n3Delay ("10ms");
  Time ds2n4Delay ("10ms");
  Time ds2n5Delay ("10ms");

  uint32_t VoipFlag = 0;
  uint32_t VtFlag = 0;
  uint32_t VoipPacket = 160;
  uint32_t VtPacket = 522;
  DataRate VoipRate ("80Kbps");
  uint32_t n0Voip = 3;
  uint32_t n1Voip = 2;
  uint32_t n2Voip = 1;
  uint32_t n1Vt = 1;

  std::string Videotrace = "../video_traces/Verbose_Jurassic.dat";

  // Command line arguments
  CommandLine cmd;
  cmd.AddValue ("SimName", "Simulation name (default: Test).", SimName);
  cmd.AddValue ("Videotrace", "Video trace file to use (default: Verbose_Jurassic.dat).", Videotrace);

  cmd.AddValue ("SimDuration", "Simulation duration (default: 60.0 sec).", SimDuration);

  cmd.AddValue ("Verbose", "Turn verbosity: [0-2] (default: 0).", Verbose);
  cmd.AddValue ("Pcap", "Turn pcap tracing: [0,1] (default: 0).", Pcap);
  cmd.AddValue ("Ascii", "Turn ascii tracing: [0,1] (default: 0).", Ascii);
  cmd.AddValue ("Checksum", "Turn packet checksum: [0,1] (default: 0).", Checksum);
  cmd.AddValue ("NetAnim", "Turn NetAnim: [0,1] (default: 0).", NetAnim);
  cmd.AddValue ("Metadata", "Turn NetAnim Metadata: [0,1] (default: 0).", Metadata);
  cmd.AddValue ("Flowmonitor", "Turn flowmonitor: [0,1] (default: 1).", Flowmonitor);

  cmd.AddValue ("DS", "Turn DiffServ architecture: [0,1] (default: 0).", DS);
  cmd.AddValue ("Policy", "Turn DiffServ policy: [0,3] (default: 0).", Policy);
  cmd.AddValue ("MdrrMode", "MDRR mode: [ALTERNATE,STRICT] (default: ALTERNATE).", MdrrMode);
  cmd.AddValue ("MTU", "Maximum Transfer Unit for the topology (default: 1500).", MTU);
  cmd.AddValue ("EFsize", "EF queue size: [>=1] (default: 100).", EFsize);
  cmd.AddValue ("AF1size", "AF1 queue size: [>=1] (default: 100).", AF1size);
  cmd.AddValue ("AF2size", "AF2 queue size: [>=1] (default: 100).", AF2size);
  cmd.AddValue ("AF3size", "AF3 queue size: [>=1] (default: 100).", AF3size);
  cmd.AddValue ("AF4size", "AF4 queue size: [>=1] (default: 100).", AF4size);
  cmd.AddValue ("BEsize", "BE queue size: [>=1] (default: 100).", BEsize);
  cmd.AddValue ("EFweight", "EF queue weight: [>=1] (default: 1).", EFweight);
  cmd.AddValue ("AF1weight", "AF1 queue weight: [>=1] (default: 1)", AF1weight);
  cmd.AddValue ("AF2weight", "AF2 queue weight: [>=1] (default: 1)", AF1weight);
  cmd.AddValue ("AF3weight", "AF3 queue weight: [>=1] (default: 1)", AF3weight);
  cmd.AddValue ("AF4weight", "AF4 queue weight: [>=1] (default: 1)", AF4weight);
  cmd.AddValue ("BEweight", "BE queue weight: [>=1] (default: 1)", BEweight);

  cmd.AddValue ("n0ds0Rate", "n0-ds0 link data rate (default: 2Mbps)", n0ds0Rate);
  cmd.AddValue ("n1ds0Rate", "n1-ds0 link data rate (default: 2Mbps)", n1ds0Rate);
  cmd.AddValue ("n2ds0Rate", "n2-ds0 link data rate (default: 2Mbps)", n2ds0Rate);
  cmd.AddValue ("ds0ds1Rate", "ds0-ds1 link data rate (default: 20Mbps)", ds0ds1Rate);
  cmd.AddValue ("ds1ds2Rate", "ds1-ds2 link data rate (default: 20Mbps)", ds0ds1Rate);
  cmd.AddValue ("ds2n3Rate", "ds2-n0 link data rate (default: 2Mbps)", ds2n3Rate);
  cmd.AddValue ("ds2n4Rate", "ds2-n1 link data rate (default: 2Mbps)", ds2n4Rate);
  cmd.AddValue ("ds2n5Rate", "ds2-n2 link data rate (default: 2Mbps)", ds2n5Rate);

  cmd.AddValue ("n0ds0Delay", "n0-ds0 link delay (default: 10ms)", n0ds0Delay);
  cmd.AddValue ("n1ds0Delay", "n1-ds0 link data rate (default: 10ms)", n1ds0Delay);
  cmd.AddValue ("n2ds0Delay", "n2-ds0 link data rate (default: 10ms)", n2ds0Delay);
  cmd.AddValue ("ds0ds1Delay", "ds0-ds1 link data rate (default: 20ms)", ds0ds1Delay);
  cmd.AddValue ("ds1ds2Delay", "ds1-ds2 link data rate (default: 20ms)", ds1ds2Delay);
  cmd.AddValue ("ds2n3Delay", "ds2-n0 link data rate (default: 10ms)", ds2n3Delay);
  cmd.AddValue ("ds2n4Delay", "ds2-n1 link data rate (default: 10ms)", ds2n4Delay);
  cmd.AddValue ("ds2n5Delay", "ds2-n2 link data rate (default: 10ms)", ds2n5Delay);

  cmd.AddValue ("VoipFlag", "Turn VoIP traffic: [0,1] (default: 0)", VoipFlag);
  cmd.AddValue ("VtFlag", "Turn Videotrace traffic: [0,1] (default: 0)", VtFlag);
  cmd.AddValue ("VoipRate", "VoIP data rate (default: 80Kbps)", VoipRate);
  cmd.AddValue ("VoipPacket", "VoIP packet size (default: 160bytes)", VoipPacket);
  cmd.AddValue ("VtPacket", "Videotrace max packet size (default: 500bytes)", VtPacket);
  cmd.AddValue ("n0Voip", "VoIP applications in Node n0 (default: 3)", n0Voip);
  cmd.AddValue ("n1Voip", "VoIP applications in Node n1 (default: 2)", n1Voip);
  cmd.AddValue ("n2Voip", "VoIP applications in Node n2 (default: 1)", n2Voip);
  cmd.AddValue ("n1Vt", "Video trace applications in Node n1 (default: 1)", n1Vt);

  cmd.Parse (argc, argv);

  // CMDs validation
  NS_ASSERT (SimDuration > 0);
  NS_ASSERT (Verbose >= 0 && Verbose <= 2);
  NS_ASSERT (Pcap >= 0 && Pcap <= 1);
  NS_ASSERT (Ascii >= 0 && Ascii <= 1);
  NS_ASSERT (Checksum >= 0 && Checksum <= 1);
  NS_ASSERT (NetAnim >= 0 && NetAnim <= 1);
  NS_ASSERT (Metadata >= 0 && Metadata <= 1);
  NS_ASSERT (Flowmonitor >= 0 && Flowmonitor <= 1);
  NS_ASSERT (DS >= 0 && DS <= 1);
  NS_ASSERT (Policy >= 0 && Policy <= 3);
  NS_ASSERT (MdrrMode == "ALTERNATE" || MdrrMode == "STRICT");
  NS_ASSERT (MTU > 0);
  NS_ASSERT (EFsize >= 1 && EFweight >= 1);
  NS_ASSERT (AF1size >= 1 && AF1weight >= 1);
  NS_ASSERT (AF2size >= 1 && AF2weight >= 1);
  NS_ASSERT (AF3size >= 1 && AF3weight >= 1);
  NS_ASSERT (AF4size >= 1 && AF4weight >= 0);
  NS_ASSERT (BEsize >= 1 && BEweight >= 1);
  NS_ASSERT (VoipFlag >= 0 && VoipFlag <= 1);
  NS_ASSERT (VtFlag >= 0 && VtFlag <= 1);
  NS_ASSERT (VoipPacket > 0);
  NS_ASSERT (VtPacket > 0);
  NS_ASSERT (n0Voip > 0 );
  NS_ASSERT (n1Voip > 0 );
  NS_ASSERT (n2Voip > 0 );
  NS_ASSERT (n1Vt > 0 );

  // Logging
  LogComponentEnableAll(LOG_PREFIX_TIME);
  LogComponentEnableAll(LOG_PREFIX_NODE);
  LogComponentEnableAll(LOG_PREFIX_LEVEL);
  //LogComponentEnableAll(LOG_PREFIX_FUNC);

  if (Verbose >= 0)
    {
      LogComponentEnable(ScriptName, LOG_LEVEL_INFO);
    }
  if (Verbose >= 1)
    {
      LogComponentEnable("PacketSink", LOG_LEVEL_INFO);
      LogComponentEnable("OnOffApplication", LOG_LEVEL_INFO);
    }
  if (Verbose >= 2)
    {
      LogComponentEnable("DiffServQueue", LOG_LEVEL_INFO);
      LogComponentEnable("DiffServ", LOG_LEVEL_INFO);
      LogComponentEnable("Mdrr", LOG_LEVEL_INFO);
      LogComponentEnable("Policy", LOG_LEVEL_INFO);
      LogComponentEnable("DiffServPointToPointHelper", LOG_LEVEL_INFO);
      LogComponentEnable("SlaHelper", LOG_LEVEL_INFO);
    }

  if (Checksum)
    {
      GlobalValue::Bind ("ChecksumEnabled", BooleanValue (true));
    }

  // Create Nodes
  NodeContainer nodes;
  nodes.Create (9);
  Ptr<Node> n0 = nodes.Get (0);
  Ptr<Node> n1 = nodes.Get (1);
  Ptr<Node> n2 = nodes.Get (2);
  Ptr<Node> ds0 = nodes.Get (3);
  Ptr<Node> ds1 = nodes.Get (4);
  Ptr<Node> ds2 = nodes.Get (5);
  Ptr<Node> n3 = nodes.Get (6);
  Ptr<Node> n4 = nodes.Get (7);
  Ptr<Node> n5 = nodes.Get (8);

  // Create Point to Point NetDevices & Channels
  // n0 - ds0
  PointToPointHelper p2p0;
  p2p0.SetDeviceAttribute ("DataRate", DataRateValue (n0ds0Rate));
  p2p0.SetChannelAttribute ("Delay", TimeValue (n0ds0Delay));
  p2p0.SetDeviceAttribute ("Mtu", UintegerValue (MTU));
  NetDeviceContainer dev0;
  dev0 = p2p0.Install (n0, ds0);
  // n1 - ds0
  PointToPointHelper p2p1;
  p2p1.SetDeviceAttribute ("DataRate", DataRateValue (n1ds0Rate));
  p2p1.SetChannelAttribute ("Delay", TimeValue (n1ds0Delay));
  p2p1.SetDeviceAttribute ("Mtu", UintegerValue (MTU));
  NetDeviceContainer dev1;
  dev1 = p2p1.Install (n1, ds0);
  // n2 - ds0
  PointToPointHelper p2p2;
  p2p2.SetDeviceAttribute ("DataRate", DataRateValue (n2ds0Rate));
  p2p2.SetChannelAttribute ("Delay", TimeValue (n2ds0Delay));
  p2p2.SetDeviceAttribute ("Mtu", UintegerValue (MTU));
  NetDeviceContainer dev2;
  dev2 = p2p2.Install (n2, ds0);

  NetDeviceContainer dev3;
  NetDeviceContainer dev4;
  DiffServPointToPointHelper p2p3;
  DiffServPointToPointHelper p2p4;
  if (DS)
    {
      // ds0 - ds1
      p2p3.SetDeviceAttribute ("DataRate", DataRateValue (ds0ds1Rate));
      p2p3.SetChannelAttribute ("Delay", TimeValue (ds0ds1Delay));
      p2p3.SetMode (DiffServQueue::EDGE);
      if (MdrrMode == "ALTERNATE")
        p2p3.SetMdrrMode (Mdrr::ALTERNATE);
      else
        p2p3.SetMdrrMode (Mdrr::STRICT);
      p2p3.SetWeights (EFweight, AF1weight, AF2weight, AF3weight, AF4weight, BEweight);
      p2p3.SetSizes (EFsize, AF1size, AF2size, AF3size, AF4size, BEsize);
      // Updates the MTU value used by DiffServQueue,
      // and changes the MTU of the PointToPoint device,
      // during Install ().
      // So, there is no need to call SetDeviceAttribute () again.
      p2p3.SetMtu (MTU);
      dev3 = p2p3.Install (ds0, ds1);
      // ds1 - ds2
      p2p4.SetDeviceAttribute ("DataRate", DataRateValue (ds1ds2Rate));
      p2p4.SetChannelAttribute ("Delay", TimeValue (ds1ds2Delay));
      p2p4.SetMode (DiffServQueue::CORE);
      if (MdrrMode == "ALTERNATE")
        p2p4.SetMdrrMode (Mdrr::ALTERNATE);
      else
        p2p4.SetMdrrMode (Mdrr::STRICT);
      p2p4.SetWeights (EFweight, AF1weight, AF2weight, AF3weight, AF4weight, BEweight);
      p2p4.SetSizes (EFsize, AF1size, AF2size, AF3size, AF4size, BEsize);
      p2p4.SetMtu (MTU);
      dev4 = p2p4.Install (ds1, ds2);
    }
  else
    {
      // ds0 - ds1
      PointToPointHelper p2p3;
      p2p3.SetDeviceAttribute ("DataRate", DataRateValue (ds0ds1Rate));
      p2p3.SetChannelAttribute ("Delay", TimeValue (ds0ds1Delay));
      dev3 = p2p3.Install (ds0, ds1);
      // ds1 - ds2
      PointToPointHelper p2p4;
      p2p4.SetDeviceAttribute ("DataRate", DataRateValue (ds1ds2Rate));
      p2p4.SetChannelAttribute ("Delay", TimeValue (ds1ds2Delay));
      dev4 = p2p4.Install (ds1, ds2);
    }

  // ds2 - n3
  PointToPointHelper p2p5;
  p2p5.SetDeviceAttribute ("DataRate", DataRateValue (ds2n3Rate));
  p2p5.SetChannelAttribute ("Delay", TimeValue (ds2n3Delay));
  p2p5.SetDeviceAttribute ("Mtu", UintegerValue (MTU));
  NetDeviceContainer dev5;
  dev5 = p2p5.Install (ds2, n3);
  // ds2 - n4
  PointToPointHelper p2p6;
  p2p6.SetDeviceAttribute ("DataRate", DataRateValue (ds2n4Rate));
  p2p6.SetChannelAttribute ("Delay", TimeValue (ds2n4Delay));
  p2p6.SetDeviceAttribute ("Mtu", UintegerValue (MTU));
  NetDeviceContainer dev6;
  dev6 = p2p6.Install (ds2, n4);
  // ds2 - n5
  PointToPointHelper p2p7;
  p2p7.SetDeviceAttribute ("DataRate", DataRateValue (ds2n5Rate));
  p2p7.SetChannelAttribute ("Delay", TimeValue (ds2n5Delay));
  p2p7.SetDeviceAttribute ("Mtu", UintegerValue (MTU));
  NetDeviceContainer dev7;
  dev7 = p2p7.Install (ds2, n5);

  // Install Internet stack
  InternetStackHelper stack;
  stack.Install (nodes);

  // Assign IPv4 addresses
  Ipv4AddressHelper ipv4;
  ipv4.SetBase ("10.1.1.0", "255.255.255.0");
  Ipv4InterfaceContainer i0 = ipv4.Assign (dev0);
  ipv4.SetBase ("10.1.2.0", "255.255.255.0");
  Ipv4InterfaceContainer i1 = ipv4.Assign (dev1);
  ipv4.SetBase ("10.1.3.0", "255.255.255.0");
  Ipv4InterfaceContainer i2 = ipv4.Assign (dev2);
  ipv4.SetBase ("10.2.1.0", "255.255.255.0");
  Ipv4InterfaceContainer i3 = ipv4.Assign (dev3);
  ipv4.SetBase ("10.2.2.0", "255.255.255.0");
  Ipv4InterfaceContainer i4 = ipv4.Assign (dev4);
  ipv4.SetBase ("10.3.1.0", "255.255.255.0");
  Ipv4InterfaceContainer i5 = ipv4.Assign (dev5);
  ipv4.SetBase ("10.3.2.0", "255.255.255.0");
  Ipv4InterfaceContainer i6 = ipv4.Assign (dev6);
  ipv4.SetBase ("10.3.3.0", "255.255.255.0");
  Ipv4InterfaceContainer i7 = ipv4.Assign (dev7);

  // Populate routing tables
  Ipv4GlobalRoutingHelper::PopulateRoutingTables ();

  // Create Applications
  uint16_t VoipPort = 1000;
  uint16_t VtPort = 2000;

  double serverStart = SimStart + 1;
  double serverStop = SimDuration;
  double clientStart = serverStart + 1;
  double clientStop = serverStop - 1;

  // Servers
  // Node n3
  // VoIP sink
  if (VoipFlag)
    {
      PacketSinkHelper n3_VoipSink ("ns3::UdpSocketFactory",
                                    InetSocketAddress (Ipv4Address::GetAny (), VoipPort));
      ApplicationContainer n3_VoipApp = n3_VoipSink.Install (n3);
      n3_VoipApp.Start (Seconds (serverStart));
      n3_VoipApp.Stop (Seconds (serverStop));
    }

  // Node n4
  // VoIP sink
  if (VoipFlag)
    {
      PacketSinkHelper n4_VoipSink ("ns3::UdpSocketFactory",
                                    InetSocketAddress (Ipv4Address::GetAny (), VoipPort));
      ApplicationContainer n4_VoipApp = n4_VoipSink.Install (n4);
      n4_VoipApp.Start (Seconds (serverStart));
      n4_VoipApp.Stop (Seconds (serverStop));
    }
  // Videotrace sink
  if (VtFlag)
    {
      PacketSinkHelper n4_VtSink ("ns3::UdpSocketFactory",
                                  InetSocketAddress (Ipv4Address::GetAny (), VtPort));
      ApplicationContainer n4_VtApp = n4_VtSink.Install (n4);
      n4_VtApp.Start (Seconds (serverStart));
      n4_VtApp.Stop (Seconds (serverStop));
    }

  // Node n5
  // VoIP sink
  if (VoipFlag)
    {
      PacketSinkHelper n5_VoipSink ("ns3::UdpSocketFactory",
                                    InetSocketAddress (Ipv4Address::GetAny (), VoipPort));
      ApplicationContainer n5_VoipApp = n5_VoipSink.Install (n5);
      n5_VoipApp.Start (Seconds (serverStart));
      n5_VoipApp.Stop (Seconds (serverStop));
    }
  // Videotrace sink
  if (VtFlag)
    {
      PacketSinkHelper n5_VtSink ("ns3::UdpSocketFactory",
                                  InetSocketAddress (Ipv4Address::GetAny (), VtPort));
      ApplicationContainer n5_VtApp = n5_VtSink.Install (n5);
      n5_VtApp.Start (Seconds (serverStart));
      n5_VtApp.Stop (Seconds (serverStop));
    }

  // Clients
  // VoIP
  if (VoipFlag)
    {
      NS_LOG_DEBUG ("Creating: " << n0Voip << " VoIP in n0.");
      std::vector<Ptr<VoIpHelper> > n0_Voip;
      for (uint32_t i = 0; i < n0Voip; ++i)
        {
          Ptr<VoIpHelper> Voip_tmp = CreateObject<VoIpHelper> ();
          Voip_tmp->SetDuration (clientStart, clientStop);
          Voip_tmp->SetDestination ("10.3.1.2", VoipPort);
          Voip_tmp->SetAttribute ("Rate", DataRateValue (VoipRate));
          Voip_tmp->SetAttribute ("PacketSize", UintegerValue (VoipPacket));
          Voip_tmp->Install (n0);
          n0_Voip.push_back (Voip_tmp);
        }

      NS_LOG_DEBUG ("Creating: " << n1Voip << " VoIP in n1.");
      std::vector<Ptr<VoIpHelper> > n1_Voip;
      for (uint32_t i = 0; i < n1Voip; ++i)
        {
          Ptr<VoIpHelper> Voip_tmp = CreateObject<VoIpHelper> ();
          Voip_tmp->SetDuration (clientStart, clientStop);
          Voip_tmp->SetDestination ("10.3.2.2", VoipPort);
          Voip_tmp->SetAttribute ("Rate", DataRateValue (VoipRate));
          Voip_tmp->SetAttribute ("PacketSize", UintegerValue (VoipPacket));
          Voip_tmp->Install (n1);
          n1_Voip.push_back (Voip_tmp);
        }

      NS_LOG_DEBUG ("Creating: " << n2Voip << " VoIP in n2.");
      std::vector<Ptr<VoIpHelper> > n2_Voip;
      for (uint32_t i = 0; i < n2Voip; ++i)
        {
          Ptr<VoIpHelper> Voip_tmp = CreateObject<VoIpHelper> ();
          Voip_tmp->SetDuration (clientStart, clientStop);
          Voip_tmp->SetDestination ("10.3.3.2", VoipPort);
          Voip_tmp->SetAttribute ("Rate", DataRateValue (VoipRate));
          Voip_tmp->SetAttribute ("PacketSize", UintegerValue (VoipPacket));
          Voip_tmp->Install (n2);
          n2_Voip.push_back (Voip_tmp);
        }
    }
  // Video streaming
  if (VtFlag)
    {
      NS_LOG_DEBUG ("Creating: " << n1Vt << " Video streaming in n1.");
      std::vector<Ptr<VideoTraceHelper> > n1_Vt;
      for (uint32_t i = 0; i < n1Vt; ++i)
        {
          Ptr<VideoTraceHelper> Vt_tmp = CreateObject<VideoTraceHelper> ();
          Vt_tmp->SetDuration (clientStart, clientStop);
          Vt_tmp->SetDestination ("10.3.2.2", VtPort);
          Vt_tmp->SetTrace (Videotrace);
          Vt_tmp->SetAttribute ("MaxPacketSize", UintegerValue (VtPacket));
          Vt_tmp->Install (n1);
          n1_Vt.push_back (Vt_tmp);
        }
    }

  // Setup DiffServ policy
  if (DS)
    {
      if (Policy > 0) // VoIP Null policer
        {
          // VoIP
          Ptr<SlaHelper> n0_SlaVoip = CreateObject<SlaHelper> ();
          n0_SlaVoip->SetType (Policy::Null);
          n0_SlaVoip->SetGreen (Policy::Dscp, Ipv4Header::EF);
          n0_SlaVoip->SetFlow ("10.1.1.1", "10.3.1.2", UDP, 0, VoipPort);
          n0_SlaVoip->Install (ds0);

          Ptr<SlaHelper> n1_SlaVoip = CreateObject<SlaHelper> ();
          n1_SlaVoip->SetType (Policy::Null);
          n1_SlaVoip->SetGreen (Policy::Dscp, Ipv4Header::EF);
          n1_SlaVoip->SetFlow ("10.1.2.1", "10.3.2.2", UDP, 0, VoipPort);
          n1_SlaVoip->Install (ds0);

          Ptr<SlaHelper> n2_SlaVoip = CreateObject<SlaHelper> ();
          n2_SlaVoip->SetType (Policy::Null);
          n2_SlaVoip->SetGreen (Policy::Dscp, Ipv4Header::EF);
          n2_SlaVoip->SetFlow ("10.1.3.1", "10.3.3.2", UDP, 0, VoipPort);
          n2_SlaVoip->Install (ds0);
        }
      if (Policy == 2)
        {
          Policy::Metrics VtMeter;
          VtMeter.CIR = 200000;
          VtMeter.CBS = 552;

          Ptr<SlaHelper> VtSla = CreateObject<SlaHelper> ();
          VtSla->SetFlow ("10.1.2.1", "10.3.2.2", UDP, 0, VtPort);
          VtSla->SetType (Policy::TokenBucket);
          VtSla->SetGreen (Policy::Dscp, Ipv4Header::AF11);
          VtSla->SetRed (Policy::Drop);
          VtSla->SetMetrics (VtMeter);

          VtSla->Install (ds0);
        }
      if (Policy == 3)
        {
          Policy::Metrics VtMeter;
          VtMeter.CIR = 100000;
          VtMeter.PIR = 150000;

          Ptr<SlaHelper> VtSla = CreateObject<SlaHelper> ();
          VtSla->SetFlow ("10.1.2.1", "10.3.2.2", UDP, 0, VtPort);
          VtSla->SetType (Policy::TSW3CM);
          VtSla->SetGreen (Policy::Dscp, Ipv4Header::AF11);
          VtSla->SetYellow (Policy::Dscp, Ipv4Header::AF21);
          VtSla->SetRed (Policy::Drop);
          VtSla->SetMetrics (VtMeter);

          VtSla->Install (ds0);
        }
    }

  // Enable tracing
  if (Pcap)
    {
      p2p0.EnablePcapAll (SimName);
      p2p1.EnablePcapAll (SimName);
      p2p2.EnablePcapAll (SimName);
      p2p3.EnablePcapAll (SimName);
      p2p4.EnablePcapAll (SimName);
      p2p5.EnablePcapAll (SimName);
      p2p6.EnablePcapAll (SimName);
      p2p7.EnablePcapAll (SimName);
    }
  if (Ascii)
    {
      AsciiTraceHelper ascii0;
      p2p0.EnableAsciiAll (ascii0.CreateFileStream (SimName+"-n0-ds0.tr"));
      AsciiTraceHelper ascii1;
      p2p1.EnableAsciiAll (ascii1.CreateFileStream (SimName+"-n1-ds0.tr"));
      AsciiTraceHelper ascii2;
      p2p2.EnableAsciiAll (ascii2.CreateFileStream (SimName+"-n2-ds0.tr"));
      AsciiTraceHelper ascii3;
      p2p3.EnableAsciiAll (ascii3.CreateFileStream (SimName+"-ds0-ds1.tr"));
      AsciiTraceHelper ascii4;
      p2p4.EnableAsciiAll (ascii4.CreateFileStream (SimName+"-ds1-ds2.tr"));
      AsciiTraceHelper ascii5;
      p2p5.EnableAsciiAll (ascii5.CreateFileStream (SimName+"-ds2-n3.tr"));
      AsciiTraceHelper ascii6;
      p2p6.EnableAsciiAll (ascii6.CreateFileStream (SimName+"-ds2-n4.tr"));
      AsciiTraceHelper ascii7;
      p2p7.EnableAsciiAll (ascii7.CreateFileStream (SimName+"-ds2-n5.tr"));
    }

  if (NetAnim)
    {
      MobilityHelper mobility;
      mobility.SetMobilityModel ("ns3::ConstantPositionMobilityModel");
      mobility.Install (nodes);

      AnimationInterface anim (SimName + "_NetAnim.xml");
      anim.SetConstantPosition (n0, 2.0, 0.0);
      anim.SetConstantPosition (n1, 0.0, 10.0);
      anim.SetConstantPosition (n2, 2.0, 20.0);
      anim.SetConstantPosition (ds0, 10.0, 10.0);
      anim.SetConstantPosition (ds1, 20.0, 10.0);
      anim.SetConstantPosition (ds2, 30.0, 10.0);
      anim.SetConstantPosition (n3, 38.0, 0.0);
      anim.SetConstantPosition (n4, 40.0, 10.0);
      anim.SetConstantPosition (n5, 38.0, 20.0);

      anim.UpdateNodeColor (ds0, 0, 255, 0);
      anim.UpdateNodeColor (ds1, 0, 255, 0);
      anim.UpdateNodeColor (ds2, 0, 255, 0);

      anim.UpdateNodeDescription (n0, "n0");
      anim.UpdateNodeDescription (n1, "n1");
      anim.UpdateNodeDescription (n2, "n2");
      anim.UpdateNodeDescription (ds0, "ds0");
      anim.UpdateNodeDescription (ds1, "ds1");
      anim.UpdateNodeDescription (ds2, "ds2");
      anim.UpdateNodeDescription (n3, "n3");
      anim.UpdateNodeDescription (n4, "n4");
      anim.UpdateNodeDescription (n5, "n5");

      if (Metadata)
        {
          anim.EnablePacketMetadata (true);
        }
    }


  FlowMonitorHelper fmh;
  Ptr<FlowMonitor> fm = fmh.InstallAll ();

  Simulator::Stop(Seconds (SimDuration));
  NS_LOG_UNCOND ("");
  NS_LOG_UNCOND ("Running simulation for: " << SimDuration << " seconds.");
  if (DS)
    {
      NS_LOG_UNCOND ("DiffServ: enabled");
      if (Policy)
        NS_LOG_UNCOND ("Policy: enabled");
      else
        NS_LOG_UNCOND ("Policy: disabled");
      NS_LOG_UNCOND ("MDRR mode: " << MdrrMode);
    }
  else
    {
      NS_LOG_UNCOND ("DiffServ: disabled");
    }
  NS_LOG_UNCOND ("");
  NS_LOG_UNCOND ("*****************");
  NS_LOG_UNCOND ("Simulation start.");
  NS_LOG_UNCOND ("*****************");
  if (Verbose > 0)
    NS_LOG_UNCOND ("Sim time|#|[Level]|Message");

  Simulator::Run ();

  if (Flowmonitor)
    {
      fm->SerializeToXmlFile (SimName + ".xml", true, true);

      fm->CheckForLostPackets ();
      Ptr<Ipv4FlowClassifier> fc = DynamicCast<Ipv4FlowClassifier> (fmh.GetClassifier ());

      std::map<FlowId, FlowMonitor::FlowStats> fs = fm->GetFlowStats ();
      std::map<FlowId, FlowMonitor::FlowStats>::const_iterator it;

      NS_LOG_UNCOND ("----------------------");
      NS_LOG_UNCOND ("Simulation statistics:");
      NS_LOG_UNCOND ("----------------------");
      for (it = fs.begin (); it != fs.end (); ++it)
        {
          NS_LOG_UNCOND ("Flow ID: " << it->first);

          NS_LOG_UNCOND ("Flow route: " << fc->FindFlow (it->first).sourceAddress
                                        << " --> " << fc->FindFlow (it->first).destinationAddress);

          NS_LOG_UNCOND ("Tx packets: " << it->second.txPackets
                                        << " (" << it->second.txBytes/1024 << " Kb)");

          NS_LOG_UNCOND ("Rx packets: " << it->second.rxPackets
                                        << " (" << it->second.rxBytes/1024 << " Kb)");

          NS_LOG_UNCOND ("Lost packets: " << it->second.txPackets - it->second.rxPackets
                                          << " (" << (it->second.txBytes -
                                             it->second.rxBytes)/1024 << " Kb)");

          NS_LOG_UNCOND ("Delay sum: " << it->second.delaySum.GetMilliSeconds () << " ms");
          NS_LOG_UNCOND ("Jitter sum: "<< it->second.jitterSum.GetMilliSeconds () << " ms");

          NS_LOG_UNCOND ("Throughput: " << (it->second.rxBytes * 8.0) /
                                        (it->second.timeLastRxPacket.GetSeconds() -
                                         it->second.timeFirstTxPacket.GetSeconds()) /
                                         1024 << " Kbps");
          NS_LOG_UNCOND ("");
        }
    }

  Simulator::Destroy ();
  NS_LOG_UNCOND ("***************");
  NS_LOG_UNCOND ("Simulation end.");
  NS_LOG_UNCOND ("***************");

  return 0;
}
