/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2013 University of Patras,
 * Department of Computer Engineering & Informatics
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * Author: Christos Nianiakas <cnianiak_at_gmail_dot_com>
 */

#ifndef __DIFFSERV_HELPER_H__
#define __DIFFSERV_HELPER_H__

#include "ns3/diffserv-queue.h"
#include "ns3/object-factory.h"
#include "ns3/net-device-container.h"
#include "ns3/node-container.h"
#include "ns3/deprecated.h"
#include "ns3/trace-helper.h"

namespace ns3 {

/**
 * \brief A DiffServ architecture implementation
 *
 * DiffServPointToPointHelper class is based in PointToPointHelper
 *
 * \brief Build a set of DiffServ PointToPointNetDevice objects
 *
 * Normally we eschew multiple inheritance, however, the classes
 * PcapUserHelperForDevice and AsciiTraceUserHelperForDevice are
 * "mixins".
 */

class DiffServPointToPointHelper : public PcapHelperForDevice, public AsciiTraceHelperForDevice
{
public:
 /**
  * Create a PointToPointHelper to make life easier when creating point to
  * point networks.
  */
  DiffServPointToPointHelper ();
  virtual ~DiffServPointToPointHelper ();
 /**
  * Each point to point net device must have a queue to pass packets through.
  * This method allows one to set the type of the queue that is automatically
  * created when the device is created and attached to a node.
  *
  * \param type the type of queue
  * \param n1 the name of the attribute to set on the queue
  * \param v1 the value of the attribute to set on the queue
  * \param n2 the name of the attribute to set on the queue
  * \param v2 the value of the attribute to set on the queue
  * \param n3 the name of the attribute to set on the queue
  * \param v3 the value of the attribute to set on the queue
  * \param n4 the name of the attribute to set on the queue
  * \param v4 the value of the attribute to set on the queue
  *
  * Set the type of queue to create and associated to each
  * PointToPointNetDevice created through PointToPointHelper::Install.
  */
  void SetQueue (std::string type,
                 std::string n1 = "", const AttributeValue &v1 = EmptyAttributeValue (),
                 std::string n2 = "", const AttributeValue &v2 = EmptyAttributeValue (),
                 std::string n3 = "", const AttributeValue &v3 = EmptyAttributeValue (),
                 std::string n4 = "", const AttributeValue &v4 = EmptyAttributeValue ());
 /**
  * Set an attribute value to be propagated to each NetDevice created by the
  * helper.
  *
  * \param name the name of the attribute to set
  * \param value the value of the attribute to set
  *
  * Set these attributes on each ns3::PointToPointNetDevice created
  * by PointToPointHelper::Install
  */
  void SetDeviceAttribute (std::string name, const AttributeValue &value);
 /**
  * Set an attribute value to be propagated to each Channel created by the
  * helper.
  *
  * \param name the name of the attribute to set
  * \param value the value of the attribute to set
  *
  * Set these attribute on each ns3::PointToPointChannel created
  * by PointToPointHelper::Install
  */
  void SetChannelAttribute (std::string name, const AttributeValue &value);
 /**
  * \param c a set of nodes
  *
  * This method creates a ns3::PointToPointChannel with the
  * attributes configured by PointToPointHelper::SetChannelAttribute,
  * then, for each node in the input container, we create a
  * ns3::PointToPointNetDevice with the requested attributes,
  * a queue for this ns3::NetDevice, and associate the resulting
  * ns3::NetDevice with the ns3::Node and ns3::PointToPointChannel.
  */
  NetDeviceContainer Install (NodeContainer c);
 /**
  * \param a first node
  * \param b second node
  *
  * Saves you from having to construct a temporary NodeContainer.
  * Also, if MPI is enabled, for distributed simulations,
  * appropriate remote point-to-point channels are created.
  */
  NetDeviceContainer Install (Ptr<Node> a, Ptr<Node> b);
 /**
  * \param a first node
  * \param bName name of second node
  *
  * Saves you from having to construct a temporary NodeContainer.
  */
  NetDeviceContainer Install (Ptr<Node> a, std::string bName);
 /**
  * \param aName Name of first node
  * \param b second node
  *
  * Saves you from having to construct a temporary NodeContainer.
  */
  NetDeviceContainer Install (std::string aName, Ptr<Node> b);
 /**
  * \param aNode Name of first node
  * \param bNode Name of second node
  *
  * Saves you from having to construct a temporary NodeContainer.
  */
  NetDeviceContainer Install (std::string aNode, std::string bNode);

// DiffServ architecure
 /**
  * \param mode DiffServ mode
  *
  * Sets the DiffServ mode
  */
  void SetMode (DiffServQueue::DsMode mode);
 /**
  * \param flow Flow of the SLA
  * \param actions Actions of the SLA
  * \param metrics Metrics of the SLA
  *
  * Sets an SLA
  */
  void SetSla (DiffServ::Flow flow,
               Policy::Actions actions,
               Policy::Metrics metrics);
 /**
  * \param mode Mdrr mode
  *
  * Sets Mdrr operetional mode
  */
  void SetMdrrMode (Mdrr::MdrrMode mode);
 /**
  * \param mtu MTU of the device
  *
  * Sets the MTU of the device
  */
  void SetMtu (uint16_t mtu);
/**
 * Set the max packets each PHB queue can hold
 *
 * \param EF Expedited Forwarding PHB queue
 * \param AF1 Assured Forwarding class 1 PHB queue
 * \param AF2 Assured Forwarding class 2 PHB queue
 * \param AF3 Assured Forwarding class 3 PHB queue
 * \param AF4 Assured Forwarding class 4 PHB queue
 * \param BE Best Effort PHB queue
 */
  void SetSizes (uint32_t EF,
                 uint32_t AF1,
                 uint32_t AF2,
                 uint32_t AF3,
                 uint32_t AF4,
                 uint32_t BE);
/**
 * Set the weight of each PHB queue
 *
 * \param EF Expedited Forwarding PHB queue
 * \param AF1 Assured Forwarding class 1 PHB queue
 * \param AF2 Assured Forwarding class 2 PHB queue
 * \param AF3 Assured Forwarding class 3 PHB queue
 * \param AF4 Assured Forwarding class 4 PHB queue
 * \param BE Best Effort PHB queue
 */
  void SetWeights (uint32_t EF,
                   uint32_t AF1,
                   uint32_t AF2,
                   uint32_t AF3,
                   uint32_t AF4,
                   uint32_t BE);
private:
 /**
  * \brief Enable pcap output the indicated net device.
  *
  * NetDevice-specific implementation mechanism for hooking the trace and
  * writing to the trace file.
  *
  * \param prefix Filename prefix to use for pcap files.
  * \param nd Net device for which you want to enable tracing.
  * \param promiscuous If true capture all possible packets available at the device.
  * \param explicitFilename Treat the prefix as an explicit filename if true
  */
  virtual void EnablePcapInternal (std::string prefix, Ptr<NetDevice> nd,
                                   bool promiscuous, bool explicitFilename);
 /**
  * \brief Enable ascii trace output on the indicated net device.
  * \internal
  *
  * NetDevice-specific implementation mechanism for hooking the trace and
  * writing to the trace file.
  *
  * \param stream The output stream object to use when logging ascii traces.
  * \param prefix Filename prefix to use for ascii trace files.
  * \param nd Net device for which you want to enable tracing.
  * \param explicitFilename Treat the prefix as an explicit filename if true
  */
  virtual void EnableAsciiInternal (
    Ptr<OutputStreamWrapper> stream,
    std::string prefix,
    Ptr<NetDevice> nd,
    bool explicitFilename);

  ObjectFactory m_queueFactory;
  ObjectFactory m_channelFactory;
  ObjectFactory m_remoteChannelFactory;
  ObjectFactory m_deviceFactory;

  // DiffServ architecure
  ObjectFactory m_queueDsFactory;
  DiffServQueue::DsMode m_DsMode;
  Mdrr::MdrrMode m_MdrrMode;
  Ptr<DiffServQueue> m_queueA;
  Ptr<DiffServQueue> m_queueB;
  uint16_t m_Mtu;
  uint32_t m_EFSize;
  uint32_t m_AF1Size;
  uint32_t m_AF2Size;
  uint32_t m_AF3Size;
  uint32_t m_AF4Size;
  uint32_t m_BESize;
  uint32_t m_EFWeight;
  uint32_t m_AF1Weight;
  uint32_t m_AF2Weight;
  uint32_t m_AF3Weight;
  uint32_t m_AF4Weight;
  uint32_t m_BEWeight;
};

} // namespace ns3

#endif /* __DIFFSERV_HELPER_H__ */
