/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2013 University of Patras,
 * Department of Computer Engineering & Informatics
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * Author: Christos Nianiakas <cnianiak_at_gmail_dot_com>
 */

#include "sla-helper.h"
#include "ns3/log.h"
#include "ns3/pointer.h"
#include "ns3/point-to-point-net-device.h"
#include <vector>

namespace ns3 {

/* ... */

NS_LOG_COMPONENT_DEFINE ("SlaHelper");

NS_OBJECT_ENSURE_REGISTERED (SlaHelper);

TypeId
SlaHelper::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::SlaHelper")
    .SetParent<Object> ()
    .AddConstructor<SlaHelper> ()
    ;
  return tid;
}

SlaHelper::SlaHelper ()
{
  flowOK = false;
  metricsOK = false;
  green = true;
  yellow = true;
  red = true;
}

SlaHelper::~SlaHelper ()
{
}

void
SlaHelper::SetFlow (Ipv4Address srcAddress,
                    Ipv4Address destAddress,
                    uint8_t protocol,
                    uint16_t srcPort,
                    uint16_t destPort)
{
  m_Flow.SrcAddress = srcAddress;
  m_Flow.DestAddress = destAddress;
  m_Flow.Protocol = protocol;
  m_Flow.SrcPort = srcPort;
  m_Flow.DestPort = destPort;
  flowOK = true;
}

void
SlaHelper::SetType (Policy::Type type)
{
  m_Actions.Type = type;
  if (type == Policy::Null)
    {
      green = false;
      metricsOK = true;
    }
  else if (type == Policy::TSW2CM)
    {
      green = false;
      red = false;
    }
  else if (type == Policy::TSW3CM)
    {
      green = false;
      yellow = false;
      red = false;
    }
  else if (type == Policy::TokenBucket)
    {
      green = false;
      red = false;
    }
  else if (type == Policy::srTCM)
    {
      green = false;
      yellow = false;
      red = false;
    }
  else if (type == Policy::trTCM)
    {
      green = false;
      yellow = false;
      red = false;
    }
}

void
SlaHelper::SetMetrics (Policy::Metrics metrics)
{
  m_Metrics = metrics;
  metricsOK = true;
}

void
SlaHelper::SetGreen (Policy::Mode mode)
{
  NS_ASSERT (mode == Policy::Drop);
  m_Actions.Green.Mode = mode;
  green = true;
}

void
SlaHelper::SetGreen (Policy::Mode mode, Ipv4Header::DscpType dscp)
{
  NS_ASSERT (mode == Policy::Dscp);
  m_Actions.Green.Mode = mode;
  m_Actions.Green.Dscp = dscp;
  green = true;
}

void
SlaHelper::SetYellow (Policy::Mode mode)
{
  NS_ASSERT (mode == Policy::Drop);
  m_Actions.Yellow.Mode = mode;
  yellow = true;
}

void
SlaHelper::SetYellow (Policy::Mode mode, Ipv4Header::DscpType dscp)
{
  NS_ASSERT (mode == Policy::Dscp);
  m_Actions.Yellow.Mode = mode;
  m_Actions.Yellow.Dscp = dscp;
  yellow = true;
}

void
SlaHelper::SetRed (Policy::Mode mode)
{
  NS_ASSERT (mode == Policy::Drop);
  m_Actions.Red.Mode = mode;
  red = true;
}

void
SlaHelper::SetRed (Policy::Mode mode, Ipv4Header::DscpType dscp)
{
  NS_ASSERT (mode == Policy::Dscp);
  m_Actions.Red.Mode = mode;
  m_Actions.Red.Dscp = dscp;
  red = true;
}

void
SlaHelper::Install (Ptr<Node> node)
{
  NS_ASSERT (green == true && yellow == true && red == true);
  NS_ASSERT (flowOK == true && metricsOK == true);

  NS_LOG_DEBUG ("Setting SLA in requested Node.");
  for (uint32_t i = 0; i < node->GetNDevices (); ++ i)
    {
      NS_LOG_DEBUG ("NetDevice #:" << i);
      Ptr<NetDevice> dev = node->GetDevice (i);
      Ptr<PointToPointNetDevice> p2pdev = dev->GetObject<PointToPointNetDevice> ();
      if (p2pdev)
        {
          PointerValue ptr;
          p2pdev->GetAttribute ("TxQueue", ptr);
          Ptr<DiffServQueue> dsq = ptr.Get<DiffServQueue> ();
          if (dsq)
            {
              dsq->SetSla (m_Flow, m_Actions, m_Metrics);
              NS_LOG_DEBUG ("SLA set.");
            }
          else
            NS_LOG_DEBUG ("Not a DiffServQueue queue. Skipping.");
        }
      else
        NS_LOG_DEBUG ("Not a point-to-point NetDevice. Skipping.");
    }
}

} //namespace ns3
