/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2013 University of Patras,
 * Department of Computer Engineering & Informatics
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * Author: Christos Nianiakas <cnianiak_at_gmail_dot_com>
 */

#include "diffserv-queue.h"
#include "ns3/log.h"
#include "ns3/enum.h"
#include "ns3/uinteger.h"
#include "ns3/ppp-header.h"
#include "ns3/tcp-header.h"
#include "ns3/udp-header.h"
#include <sstream>

namespace ns3 {

/* ... */

NS_LOG_COMPONENT_DEFINE ("DiffServQueue");
NS_OBJECT_ENSURE_REGISTERED (DiffServQueue);

TypeId
DiffServQueue::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::DiffServQueue")
    .SetParent<Queue> ()
    .AddConstructor<DiffServQueue> ()
    .AddAttribute ("MODE", "DiffServ Node type (CORE,EDGE).",
                   EnumValue (CORE),
                   MakeEnumAccessor (&DiffServQueue::m_DsMode),
                   MakeEnumChecker (CORE, "CORE",
                                    EDGE, "EDGE"))
    .AddAttribute ("MTU", "MTU of the interaface.",
                   UintegerValue (1500),
                   MakeUintegerAccessor (&DiffServQueue::m_Mtu),
                   MakeUintegerChecker<uint16_t> ())
    ;
  return tid;
}

// Constructor
DiffServQueue::DiffServQueue ()
{
  NS_LOG_FUNCTION_NOARGS ();

  m_DiffServ = CreateObject<DiffServ> ();
  m_Mdrr = CreateObject<Mdrr> ();
  m_Mdrr->SetDiffServ (m_DiffServ);
  m_MdrrReady = false;
  m_SlaCtr = 0;
}

// Destructor
DiffServQueue::~DiffServQueue ()
{
  NS_LOG_FUNCTION_NOARGS ();
}

// DoEnqueue
bool
DiffServQueue::DoEnqueue (Ptr<Packet> p)
{
  NS_LOG_FUNCTION (this << p);

  NS_LOG_DEBUG ("----------");
  NS_LOG_DEBUG ("DoEnqueue.");
  NS_LOG_DEBUG ("----------");

  // Boundary DiffServ Node.
  if (m_DsMode == EDGE)
    {
      NS_LOG_DEBUG ("DiffServ EDGE.");
      // Multi-field classifier.
      DiffServ::Flow flow;
      flow = m_DiffServ->MfClassifier (p);

      // Check SLA database for packet's flow.
      DiffServQueue::Sla sla = GetSla (flow);
      Ipv4Header ipv4Header;
      if (flow.status && sla.status)
        {
          NS_LOG_DEBUG ("Classified packet as DSCP: "
                        << ipv4Header.DscpTypeToString (sla.Flow.Dscp));
        }
      else
        {
          NS_LOG_DEBUG ("Failed to classify packet.");
          NS_LOG_DEBUG ("Assuming packet DSCP: Default.");
          sla.Flow.Dscp = Ipv4Header::DscpDefault;
        }

        // SLA for flow found.
        if (sla.status)
        {
          // Meter packet and apply SLA's action.
          NS_LOG_DEBUG ("Checking conformance of packet.");
          Policy::Conformance conformance = sla.FlowPolicy->Meter (p);
          Policy::Actions actions = sla.FlowPolicy->GetActions ();

          // Conforming packet.
          if (conformance == Policy::Green)
            {
              NS_LOG_DEBUG ("Packet is: Green.");
              if (actions.Green.Mode == Policy::Dscp)
                {
                  sla.Flow.Dscp = actions.Green.Dscp;
                  NS_LOG_INFO ("DSCP action: " << ipv4Header.DscpTypeToString (sla.Flow.Dscp));
                }
              else
                {
                  NS_LOG_INFO ("Drop action.");
                  Drop (p);
                  NS_LOG_DEBUG ("Packet dropped.");
                  return false;
                }
            }
          // Non-conforming packet level 1.
          else if (conformance == Policy::Yellow)
            {
              NS_LOG_DEBUG ("Packet is: Yellow.");
              if (actions.Yellow.Mode == Policy::Dscp)
                {
                  sla.Flow.Dscp = actions.Yellow.Dscp;
                  NS_LOG_INFO ("DSCP action: " << ipv4Header.DscpTypeToString (sla.Flow.Dscp));
                }
              else
                {
                  NS_LOG_INFO ("Drop action.");
                  Drop (p);
                  NS_LOG_DEBUG ("Packet dropped.");
                  return false;
                }
            }
          // Non-conforming packet level 2.
          else if (conformance == Policy::Red)
            {
              NS_LOG_DEBUG ("Packet is: Red.");
              if (actions.Red.Mode == Policy::Dscp)
                {
                  sla.Flow.Dscp = actions.Red.Dscp;
                  NS_LOG_INFO ("DSCP action: " << ipv4Header.DscpTypeToString (sla.Flow.Dscp));
                }
              else
                {
                  NS_LOG_INFO ("Drop action.");
                  Drop (p);
                  NS_LOG_DEBUG ("Packet dropped.");
                  return false;
                }
            }
        }

      // Marker.
      bool check = m_DiffServ->Marker (p, sla.Flow.Dscp);
      if (check)
        {
          NS_LOG_DEBUG ("Marked packet with DSCP: "
                        << ipv4Header.DscpTypeToString (sla.Flow.Dscp));
        }
      else
        {
          NS_LOG_DEBUG ("Failed to mark packet.");
          Drop (p);
          NS_LOG_DEBUG ("Packet dropped.");
          return false;
        }
    } // EDGE

  // Behaviour aggregate classifier.
  //
  // The only mechanism excecuted,
  // by the DiffServ interial Nodes.
  if (m_DsMode == CORE)
    {
      NS_LOG_DEBUG ("DiffServ CORE.");
    }
  bool status = m_DiffServ->BaClassifier (p);
  if (status)
    {
      NS_LOG_DEBUG ("Packet enqueued.");
      return true;
    }
  else
    {
      NS_LOG_DEBUG ("Failed to enqueue packet.");
      Drop (p);
      NS_LOG_DEBUG ("Packet dropped.");
      return false;
    }
}

// DoDequeue
Ptr<Packet>
DiffServQueue::DoDequeue (void)
{
  NS_LOG_FUNCTION (this);
  NS_LOG_DEBUG ("----------");
  NS_LOG_DEBUG ("DoDequeue.");
  NS_LOG_DEBUG ("----------");

  // Manage the dequeue proccess
  // with MDRR scheduler.
  NS_LOG_DEBUG (">>MDRR scheduler<<");
  // Initialize MDRR scheduler,
  // (runs only once).
  if (!m_MdrrReady)
    {
      // Set MDRR operational mode.
      m_Mdrr->SetAttribute ("MdrrMode", EnumValue (m_MdrrMode));
      // Set MTU of the interface,
      // (used in calculating the quantum value for each queue).
      m_Mdrr->SetAttribute ("MTU", UintegerValue (m_Mtu));
      m_MdrrReady = true;
    }
  // Schedule next packet.
  Ptr<Packet> p = m_Mdrr->MdrrSchedule ();
  if (p != 0)
    NS_LOG_LOGIC ("Popped " << p);
  else
    NS_LOG_DEBUG ("Nothing to dequeue.");

  return p;
}

// DoPeek
//
// (Currently not implemented).
Ptr<const Packet>
DiffServQueue::DoPeek (void) const
{
  NS_LOG_FUNCTION (this);

  Ptr<Packet> p;

  NS_LOG_DEBUG ("Peeked packet.");

  return p;
}

// GetSla
//
// NOTES:
// Currently, only destination port is considered valid.
// Source port is set to 0, during the SLA setup.
//
// SLA database is implemented with a vector.
// This could cause performance issues,
// if the database grows large enough.
//
// A more efficient data structure can be used,
// (like an std::map with a uniqueue hash per flow for key).
DiffServQueue::Sla
DiffServQueue::GetSla (DiffServ::Flow flow)
{
  // Check SLA vector for current flow.
  std::vector<DiffServQueue::Sla>::iterator it;
  for (it = m_Sla.begin (); it != m_Sla.end (); ++it)
    {
      if (flow.SrcAddress == it->Flow.SrcAddress &&
          flow.DestAddress == it->Flow.DestAddress &&
          //flow.Dscp == it->Flow.Dscp &&
          flow.Protocol == it->Flow.Protocol &&
          //flow.SrcPort == it->Flow.SrcPort &&
          flow.DestPort == it->Flow.DestPort)
        {
          NS_LOG_DEBUG ("Found flow SLA, with ID: "<< it->SlaId);
          return *it;
        }
    }

  NS_LOG_DEBUG ("SLA for flow, not found.");
  DiffServQueue::Sla sla;
  sla.status = false;
  return sla;
}

// SetSla
// Insert a new SLA into database.
void
DiffServQueue::SetSla (DiffServ::Flow flow,
                       Policy::Actions actions,
                       Policy::Metrics metrics)
{
  DiffServQueue::Sla sla;
  m_SlaCtr += 1;
  sla.SlaId = m_SlaCtr;
  sla.Flow = flow;
  sla.Flow.Dscp = actions.Green.Dscp;
  sla.status = true;

  Ptr<Policy> policy = CreateObject<Policy> ();
  sla.FlowPolicy = policy;
  sla.FlowPolicy->SetPolicy (actions, metrics);

  m_Sla.push_back (sla);
}


// SetMdrrMode
// Set MDDR operational mode.
void
DiffServQueue::SetMdrrMode (Mdrr::MdrrMode mode)
{
  m_MdrrMode = mode;
}

// SetSizes
// Set max size for available queues.
void
DiffServQueue::SetSizes (uint32_t EF,
                         uint32_t AF1,
                         uint32_t AF2,
                         uint32_t AF3,
                         uint32_t AF4,
                         uint32_t BE)
{
  m_DiffServ->SetSizes (EF, AF1, AF2, AF3, AF4, BE);
}

// SetWeights
// Set MDRR's weight's for each queue.
void
DiffServQueue::SetWeights (uint32_t EF,
                           uint32_t AF1,
                           uint32_t AF2,
                           uint32_t AF3,
                           uint32_t AF4,
                           uint32_t BE)
{
  m_Mdrr->SetWeights (EF, AF1, AF2, AF3, AF4, BE);
}

} //namespace ns3
