/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2013 University of Patras,
 * Department of Computer Engineering & Informatics
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * Author: Christos Nianiakas <cnianiak_at_gmail_dot_com>
 */

#ifndef __DIFFSERV_QUEUE_H__
#define __DIFFSERV_QUEUE_H__

#include "diffserv.h"
#include "mdrr.h"
#include "policy.h"
#include "ns3/queue.h"
#include "ns3/packet.h"
#include "ns3/ipv4-header.h"
#include <vector>

namespace ns3 {

/**
 * \brief A DiffServ architecture implementation
 *
 * DiffServQueue class is the main entry point for the architecture
 * Currently, can be installed as the Tx queue of point-to-point Net Devices
 */

class DiffServQueue : public Queue
{
public:
  static TypeId GetTypeId (void);

/**
 * \brief DiffServQueue Constructor
 *
 * Creates a DiffServQueue instance into CORE mode by default
 */
  DiffServQueue ();
  virtual ~DiffServQueue ();

  // DiffServ operation modes
  enum DsMode
    {
      CORE,
      EDGE
    };

  // DiffServ SLA
  struct Sla
    {
      bool status;
      uint32_t SlaId;
      DiffServ::Flow Flow;
      Ptr<Policy> FlowPolicy;
    };

/**
 * Search SLA database for current flow
 *
 * \param flow The flow to search
 */
  DiffServQueue::Sla GetSla (DiffServ::Flow flow);
/**
 * Set an SLA into database with the parameters provided
 *
 * \param flow The flow to assign the SLA
 * \param actions Policy mechanism to use and actions to take
 *        (max available conformance levels: Green, Yellow, Red)
 * \param metrics Metrics used by selected policy mechanism
 */
  void SetSla (DiffServ::Flow flow,
               Policy::Actions actions,
               Policy::Metrics metrics);
/**
 * Set Mdrr mode (available modes: ALTERNATE, STRICT)
 *
 * \param mode Mdrr mode to set
 */
  void SetMdrrMode (Mdrr::MdrrMode mode);
/**
 * Set the max packets each PHB queue can hold
 *
 * \param EF Expedited Forwarding PHB queue
 * \param AF1 Assured Forwarding class 1 PHB queue
 * \param AF2 Assured Forwarding class 2 PHB queue
 * \param AF3 Assured Forwarding class 3 PHB queue
 * \param AF4 Assured Forwarding class 4 PHB queue
 * \param BE Best Effort PHB queue
 */
  void SetSizes (uint32_t EF,
                 uint32_t AF1,
                 uint32_t AF2,
                 uint32_t AF3,
                 uint32_t AF4,
                 uint32_t BE);
/**
 * Set the weight of each PHB queue
 *
 * \param EF Expedited Forwarding PHB queue
 * \param AF1 Assured Forwarding class 1 PHB queue
 * \param AF2 Assured Forwarding class 2 PHB queue
 * \param AF3 Assured Forwarding class 3 PHB queue
 * \param AF4 Assured Forwarding class 4 PHB queue
 * \param BE Best Effort PHB queue
 */
  void SetWeights (uint32_t EF,
                   uint32_t AF1,
                   uint32_t AF2,
                   uint32_t AF3,
                   uint32_t AF4,
                   uint32_t BE);

private:
  // Inherited Queue virutal methods
  virtual bool DoEnqueue (Ptr<Packet> p);
  virtual Ptr<Packet> DoDequeue (void);
  virtual Ptr<const Packet> DoPeek (void) const;

  DsMode m_DsMode; // DiffServ operation mode

  uint32_t m_SlaCtr; // The number of SLA'S that are holded
  std::vector<DiffServQueue::Sla> m_Sla; // Vector to hold the SLA database

  Ptr<DiffServ> m_DiffServ; // Smart pointer to a DiffServ class instance

  Ptr<Mdrr> m_Mdrr; // Smart pointer to a Mdrr class instance
  bool m_MdrrReady; // Flag to initialize Mdrr once
  uint16_t m_Mtu; // MTU of the interface
  Mdrr::MdrrMode m_MdrrMode; // Mdrr operational mode
};

} //namespace ns3

#endif /* __DIFFSERV_QUEUE_H__ */
