/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2013 University of Patras,
 * Department of Computer Engineering & Informatics
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * Author: Christos Nianiakas <cnianiak_at_gmail_dot_com>
 */

#include "diffserv.h"
#include "ns3/log.h"
#include "ns3/enum.h"
#include "ns3/uinteger.h"
#include "ns3/ppp-header.h"
#include "ns3/tcp-header.h"
#include "ns3/udp-header.h"

namespace ns3 {

/* ... */

NS_LOG_COMPONENT_DEFINE ("DiffServ");
NS_OBJECT_ENSURE_REGISTERED (DiffServ);

TypeId
DiffServ::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::DiffServ")
    .SetParent<Object> ()
    .AddConstructor<DiffServ> ()
    .AddAttribute ("EFSize", "EF queue size (default: 100).",
                   UintegerValue (100),
                   MakeUintegerAccessor (&DiffServ::m_EFSize),
                   MakeUintegerChecker<uint32_t> ())
    .AddAttribute ("AF1Size", "AF1 queue size (default: 100).",
                   UintegerValue (100),
                   MakeUintegerAccessor (&DiffServ::m_AF1Size),
                   MakeUintegerChecker<uint32_t> ())
    .AddAttribute ("AF2Size", "AF2 queue size (default: 100).",
                   UintegerValue (100),
                   MakeUintegerAccessor (&DiffServ::m_AF2Size),
                   MakeUintegerChecker<uint32_t> ())
    .AddAttribute ("AF3Size", "AF3 queue size (default: 100).",
                   UintegerValue (100),
                   MakeUintegerAccessor (&DiffServ::m_AF3Size),
                   MakeUintegerChecker<uint32_t> ())
    .AddAttribute ("AF4Size", "AF4 queue size (default: 100).",
                   UintegerValue (100),
                   MakeUintegerAccessor (&DiffServ::m_AF4Size),
                   MakeUintegerChecker<uint32_t> ())
    .AddAttribute ("BESize", "BE queue size (default: 100).",
                   UintegerValue (100),
                   MakeUintegerAccessor (&DiffServ::m_BESize),
                   MakeUintegerChecker<uint32_t> ())
    ;
  return tid;
}

// Constructor
DiffServ::DiffServ ()
{
  NS_LOG_FUNCTION_NOARGS ();
}

// Destructor
DiffServ::~DiffServ ()
{
  NS_LOG_FUNCTION_NOARGS ();
}

// Ipv4Address
// Converts an IPv4 address to string.
std::string
DiffServ::Ipv4AddressStr (Ipv4Address address)
{
  std::ostringstream stream;
  address.Print (stream);
  std::string out = stream.str ();
  return out;
}

// MF Classifier
// Muli-field classifier.
DiffServ::Flow
DiffServ::MfClassifier (Ptr<Packet> p)
{
  NS_LOG_DEBUG (">>MF Classifier<<");

  // Create a copy of the packet,
  // to extract its headers.
  Ptr<Packet> q = p->Copy ();

  DiffServ::Flow flow;
  flow.status = false;

  // Remove Data Link layer header.
  //
  // Currently: Only PPP header is suppported.
  // If an other header is encountered,
  // classification will stop and return.
  uint32_t bytesRead = 0;
  PppHeader pppHeader;
  bytesRead = q->RemoveHeader (pppHeader);
  if (bytesRead > 0)
    {
      NS_LOG_DEBUG ("PPP Header found.");
    }
  else
    {
      NS_LOG_DEBUG ("Failed to read PPP header.");
      return flow;
    }

  // Get IP header's information.
  //
  // Currently: Only IPv4 header is supported.
  // If IPv6 is encountered,
  // classification will stop and return.
  Ipv4Header ipv4Header;
  bytesRead = q->RemoveHeader (ipv4Header);
  if (bytesRead > 0)
    {
      NS_LOG_DEBUG ("IPv4 Header found.");
      flow.SrcAddress = ipv4Header.GetSource ();
      NS_LOG_DEBUG ("IPv4 source: " << Ipv4AddressStr (flow.SrcAddress));
      flow.DestAddress = ipv4Header.GetDestination ();
      NS_LOG_DEBUG ("IPv4 destination: " << Ipv4AddressStr (flow.DestAddress));
      //flow.Dscp = ipv4Header.GetDscp ();
      //NS_LOG_DEBUG ("DSCP found: " << ipv4Header.DscpTypeToString (flow.Dscp));
      flow.Protocol = ipv4Header.GetProtocol ();
      NS_LOG_DEBUG ("IPv4 protocol: " << int(flow.Protocol));
    }
  else
    {
      NS_LOG_DEBUG ("Failed to read IPv4 header.");
      return flow;
    }

  // Get transport layer header.
  //
  // Currently: TCP and UDP are supported.
  if (flow.Protocol == TCP)
    {
      NS_LOG_DEBUG ("TCP protocol found.");
      TcpHeader tcpHeader;
      bytesRead = q->RemoveHeader (tcpHeader);
      if (bytesRead > 0)
        {
          NS_LOG_DEBUG ("TCP Header found.");
          flow.SrcPort = tcpHeader.GetSourcePort ();
          NS_LOG_DEBUG ("TCP source port: " << flow.SrcPort);
          flow.DestPort = tcpHeader.GetDestinationPort ();
          NS_LOG_DEBUG ("TCP destination port: " << flow.DestPort);
        }
      else
        {
          NS_LOG_DEBUG ("Failed to read TCP header.");
          return flow;
        }
    }
  else if (flow.Protocol == UDP)
    {
      UdpHeader udpHeader;
      bytesRead = q->RemoveHeader (udpHeader);
      if (bytesRead > 0)
        {
          NS_LOG_DEBUG ("UDP protocol found.");
          NS_LOG_DEBUG ("UDP Header found.");
          flow.SrcPort = udpHeader.GetSourcePort ();
          NS_LOG_DEBUG ("UDP source port: " << flow.SrcPort);
          flow.DestPort = udpHeader.GetDestinationPort ();
          NS_LOG_DEBUG ("UDP destination port: " << flow.DestPort);
        }
      else
        {
          NS_LOG_DEBUG ("Failed to read UDP header.");
          return flow;
        }
    }

  flow.status = true;
  return flow;
}

// Marker
bool
DiffServ::Marker (Ptr<Packet> p, Ipv4Header::DscpType dscp)
{
  NS_LOG_DEBUG (">>Marker<<");

  // Create a copy of the packet,
  // to extract its headers.
  Ptr<Packet> q = p->Copy ();

  uint32_t bytesRead = 0;
  PppHeader pppHeaderNew;
  bytesRead = q->RemoveHeader (pppHeaderNew);
  if (bytesRead > 0)
    {
      NS_LOG_DEBUG ("PPP Header found.");
    }
  else
    {
      NS_LOG_DEBUG ("Failed to read PPP header.");
      return false;
    }

  // Extract IPv4 header and mark its DSCP field.
  Ipv4Header ipv4HeaderNew;
  bytesRead = q->RemoveHeader (ipv4HeaderNew);
  if (bytesRead > 0)
    {
      NS_LOG_DEBUG ("IPv4 Header found.");
      NS_LOG_DEBUG ("Marking packet.");
      ipv4HeaderNew.SetDscp (dscp);
    }
  else
    {
      NS_LOG_DEBUG ("Failed to read IPv4 header.");
      return false;
    }

  // From the original packet:
  // Remove PPP header.
  PppHeader pppHeaderOld;
  p->RemoveHeader (pppHeaderOld);
  // Remove IPv4 header.
  Ipv4Header ipv4HeaderOld;
  p->RemoveHeader (ipv4HeaderOld);

  // Insert modified IPv4 header to it.
  p->AddHeader (ipv4HeaderNew);
  // Insert back PPP header.
  p->AddHeader (pppHeaderNew);

  return true;
}

// BA Classifier
// Behaviour aggregate classifier.
bool
DiffServ::BaClassifier (Ptr<Packet> p)
{
  NS_LOG_DEBUG (">>BA Classifier<<");

  Ptr<Packet> q = p->Copy ();

  uint32_t bytesRead = 0;
  PppHeader pppHeader;
  bytesRead = q->RemoveHeader (pppHeader);
  if (bytesRead > 0)
    {
      NS_LOG_DEBUG ("PPP Header found.");
    }
  else
    {
      NS_LOG_DEBUG ("Failed to read PPP header.");
      return false;
    }

  // Check packet's DSCP field.
  Ipv4Header::DscpType dscp;
  Ipv4Header ipv4Header;
  bytesRead = q->RemoveHeader (ipv4Header);
  if (bytesRead > 0)
    {
      NS_LOG_DEBUG ("IPv4 Header found.");
      dscp = ipv4Header.GetDscp ();
    }
  else
    {
      NS_LOG_DEBUG ("Failed to read IPv4 header.");
      return false;
    }

  // Enqueue the packet according to its DCSP.
  bool status = false;
  if (dscp == Ipv4Header::EF)
    {
      status = EFEnqueue (p);
    }
  else if (dscp == Ipv4Header::AF11 ||
           dscp == Ipv4Header::AF12 ||
           dscp == Ipv4Header::AF13)
    {
      status = AF1Enqueue (p);
    }
  else if (dscp == Ipv4Header::AF21 ||
           dscp == Ipv4Header::AF22 ||
           dscp == Ipv4Header::AF23)
    {
      status = AF2Enqueue (p);
    }
  else if (dscp == Ipv4Header::AF31 ||
           dscp == Ipv4Header::AF32 ||
           dscp == Ipv4Header::AF33)
    {
      status = AF2Enqueue (p);
    }
  else if (dscp == Ipv4Header::AF41 ||
           dscp == Ipv4Header::AF42 ||
           dscp == Ipv4Header::AF43)
    {
      status = AF4Enqueue (p);
    }
  else if (dscp == Ipv4Header::DscpDefault)
    {
      status = BEEnqueue (p);
    }

  return status;
}

// Enqueue
bool
DiffServ::EFEnqueue (Ptr<Packet> p)
{
  bool check;
  if (m_EFQueue.size () < m_EFSize)
    {
      m_EFQueue.push (p);
      NS_LOG_DEBUG ("Pushed to EF queue.");
      check = true;
    }
  else
    {
      NS_LOG_DEBUG ("EF queue full.");
      check = false;
    }

  return check;
}

bool
DiffServ::AF1Enqueue (Ptr<Packet> p)
{
  bool check;
  if (m_AF1Queue.size () < m_AF1Size)
    {
      m_AF1Queue.push (p);
      NS_LOG_DEBUG ("Pushed to AF1 queue.");
      check = true;
    }
  else
    {
      NS_LOG_DEBUG ("AF1 queue full.");
      check = false;
    }

  return check;
}

bool
DiffServ::AF2Enqueue (Ptr<Packet> p)
{
  bool check;
  if (m_AF2Queue.size () < m_AF2Size)
    {
      m_AF2Queue.push (p);
      NS_LOG_DEBUG ("Pushed to AF2 queue.");
      check = true;
    }
  else
    {
      NS_LOG_DEBUG ("AF2 queue full.");
      check = false;
    }

  return check;
}

bool
DiffServ::AF3Enqueue (Ptr<Packet> p)
{
  bool check;
  if (m_AF3Queue.size () < m_AF3Size)
    {
      m_AF3Queue.push (p);
      NS_LOG_DEBUG ("Pushed to AF3 queue.");
      check = true;
    }
  else
    {
      NS_LOG_DEBUG ("AF3 queue full.");
      check = false;
    }

  return check;
}

bool
DiffServ::AF4Enqueue (Ptr<Packet> p)
{
  bool check;
  if (m_AF4Queue.size () < m_AF4Size)
    {
      m_AF4Queue.push (p);
      NS_LOG_DEBUG ("Pushed to AF4 queue.");
      check = true;
    }
  else
    {
      NS_LOG_DEBUG ("AF4 queue full.");
      check = false;
    }

  return check;
}

bool
DiffServ::BEEnqueue (Ptr<Packet> p)
{
  bool check;
  if (m_BEQueue.size () < m_BESize)
    {
      m_BEQueue.push (p);
      NS_LOG_DEBUG ("Pushed to BE queue.");
      check = true;
    }
  else
    {
      NS_LOG_DEBUG ("BE queue full.");
      check = false;
    }

  return check;
}

// Dequeue
Ptr<Packet>
DiffServ::EFDequeue (void)
{
  Ptr<Packet> p;
  if (m_EFQueue.size () > 0)
    {
      p = m_EFQueue.front ();
      m_EFQueue.pop ();
      NS_LOG_DEBUG ("Popped from EF queue.");
    }
  else
    {
      NS_LOG_DEBUG ("EF queue empty.");
      p = 0;
    }

  return p;
}

Ptr<Packet>
DiffServ::AF1Dequeue (void)
{
  Ptr<Packet> p;
  if (m_AF1Queue.size () > 0)
    {
      p = m_AF1Queue.front ();
      m_AF1Queue.pop ();
      NS_LOG_DEBUG ("Popped from AF1 queue.");
    }
  else
    {
      NS_LOG_DEBUG ("AF1 queue empty.");
      p = 0;
    }

  return p;
}

Ptr<Packet>
DiffServ::AF2Dequeue (void)
{
  Ptr<Packet> p;
  if (m_AF2Queue.size () > 0)
    {
      p = m_AF2Queue.front ();
      m_AF2Queue.pop ();
      NS_LOG_DEBUG ("Popped from AF2 queue.");
    }
  else
    {
      NS_LOG_DEBUG ("AF2 queue empty.");
      p = 0;
    }

  return p;
}

Ptr<Packet>
DiffServ::AF3Dequeue (void)
{
  Ptr<Packet> p;
  if (m_AF3Queue.size () > 0)
    {
      p = m_AF3Queue.front ();
      m_AF3Queue.pop ();
      NS_LOG_DEBUG ("Popped from AF3 queue.");
    }
  else
    {
      NS_LOG_DEBUG ("AF3 queue empty.");
      p = 0;
    }

  return p;
}

Ptr<Packet>
DiffServ::AF4Dequeue (void)
{
  Ptr<Packet> p;
  if (m_AF4Queue.size () > 0)
    {
      p = m_AF4Queue.front ();
      m_AF4Queue.pop ();
      NS_LOG_DEBUG ("Popped from AF4 queue.");
    }
  else
    {
      NS_LOG_DEBUG ("AF4 queue empty.");
      p = 0;
    }

  return p;
}

Ptr<Packet>
DiffServ::BEDequeue (void)
{
  Ptr<Packet> p;
  if (m_BEQueue.size () > 0)
    {
      p = m_BEQueue.front ();
      m_BEQueue.pop ();
      NS_LOG_DEBUG ("Popped from BE queue.");
    }
  else
    {
      NS_LOG_DEBUG ("BE queue empty.");
      p = 0;
    }

  return p;
}

// Empty check
bool
DiffServ::EmptyEF (void)
{
  bool check;
  if (m_EFQueue.empty ())
    check = true;
  else
    check = false;

  return check;
}

bool
DiffServ::EmptyAF1 (void)
{
  bool check;
  if (m_AF1Queue.empty ())
    check = true;
  else
    check = false;

  return check;
}

bool
DiffServ::EmptyAF2 (void)
{
  bool check;
  if (m_AF2Queue.empty ())
    check = true;
  else
    check = false;

  return check;
}

bool
DiffServ::EmptyAF3 (void)
{
  bool check;
  if (m_AF3Queue.empty ())
    check = true;
  else
    check = false;

  return check;
}

bool
DiffServ::EmptyAF4 (void)
{
  bool check;
  if (m_AF4Queue.empty ())
    check = true;
  else
    check = false;

  return check;
}

bool
DiffServ::EmptyBE (void)
{
  bool check;
  if (m_BEQueue.empty ())
    check = true;
  else
    check = false;

  return check;
}

void
DiffServ::SetSizes (uint32_t EF,
                   uint32_t AF1,
                   uint32_t AF2,
                   uint32_t AF3,
                   uint32_t AF4,
                   uint32_t BE)
{
  m_EFSize = EF;
  m_AF1Size = AF1;
  m_AF2Size = AF2;
  m_AF3Size = AF3;
  m_AF4Size = AF4;
  m_BESize = BE;
}

} //namespace ns3
