/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2013 University of Patras,
 * Department of Computer Engineering & Informatics
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * Author: Christos Nianiakas <cnianiak_at_gmail_dot_com>
 */

#ifndef __DIFFSERV_H__
#define __DIFFSERV_H__

#include "ns3/object.h"
#include "ns3/packet.h"
#include "ns3/ipv4-header.h"
#include <queue>

namespace ns3 {

// Common transport protocol values used
const uint8_t TCP = 6;
const uint8_t UDP = 17;

// DiffServ supported queues
enum DsQueues
  {
    EF,
    AF1,
    AF2,
    AF3,
    AF4,
    BE
  };

/**
 * \brief A DiffServ architecture implementation
 *
 * DiffServ class holds the Tx queues defined in DiffServ architecture
 * and implements essential mechanisms to manage them
 */

class DiffServ : public Object
{
public:
  static TypeId GetTypeId (void);

/**
 * \brief DiffServ Constructor
 *
 * Creates a DiffServ instance with queue size: 100
 * for each supported queue
 */
  DiffServ ();
  virtual ~DiffServ ();

  // Holds flow information
  struct Flow
    {
      bool status;
      Ipv4Address SrcAddress;
      Ipv4Address DestAddress;
      Ipv4Header::DscpType Dscp;
      uint8_t Protocol;
      uint16_t SrcPort;
      uint16_t DestPort;
    };

/**
 * Convert Ipv4Address to string
 *
 * \param address Ipv4Address to convert
 */
  std::string Ipv4AddressStr (Ipv4Address address);
/**
 * A Multi-field classifier
 *
 * \param p Packet to classify
 *
 * \return Returns a DiffServ::Flow struct
 */
  DiffServ::Flow MfClassifier (Ptr<Packet> p);
/**
 * A Marker
 *
 * \param p Packet to mark
 * \param dscp Dscp value to mark packet with
 *
 * \return Return true in successful marking (false otherwise)
 */
  bool Marker (Ptr<Packet> p, Ipv4Header::DscpType dscp);
/**
 * A Behaviour Aggregate classifier
 *
 * \param p Packet to classify
 */
  bool BaClassifier (Ptr<Packet> p);

/**
 * Helper methods for the  PHB queues
 */
  bool EFEnqueue (Ptr<Packet>);
  bool AF1Enqueue (Ptr<Packet>);
  bool AF2Enqueue (Ptr<Packet>);
  bool AF3Enqueue (Ptr<Packet>);
  bool AF4Enqueue (Ptr<Packet>);
  bool BEEnqueue (Ptr<Packet>);

  Ptr<Packet> EFDequeue (void);
  Ptr<Packet> AF1Dequeue (void);
  Ptr<Packet> AF2Dequeue (void);
  Ptr<Packet> AF3Dequeue (void);
  Ptr<Packet> AF4Dequeue (void);
  Ptr<Packet> BEDequeue (void);

  bool EmptyEF (void);
  bool EmptyAF1 (void);
  bool EmptyAF2 (void);
  bool EmptyAF3 (void);
  bool EmptyAF4 (void);
  bool EmptyBE (void);

  void SetSizes (uint32_t EF,
                 uint32_t AF1,
                 uint32_t AF2,
                 uint32_t AF3,
                 uint32_t AF4,
                 uint32_t BE);

private:
  // Expedited Forwarding Queue
  std::queue<Ptr<Packet> > m_EFQueue;

  // Assured Forwarding Queues
  std::queue<Ptr<Packet> > m_AF1Queue;
  std::queue<Ptr<Packet> > m_AF2Queue;
  std::queue<Ptr<Packet> > m_AF3Queue;
  std::queue<Ptr<Packet> > m_AF4Queue;

  // Best Efford Queue
  std::queue<Ptr<Packet> > m_BEQueue;

  // Queue capacities
  uint32_t m_EFSize;
  uint32_t m_AF1Size;
  uint32_t m_AF2Size;
  uint32_t m_AF3Size;
  uint32_t m_AF4Size;
  uint32_t m_BESize;
};

} //namespace ns3

#endif /* __DIFFSERV_H__ */
