/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2013 University of Patras,
 * Department of Computer Engineering & Informatics
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * Author: Christos Nianiakas <cnianiak_at_gmail_dot_com>
 */

#include "mdrr.h"
#include "ns3/log.h"
#include "ns3/enum.h"
#include "ns3/uinteger.h"

NS_LOG_COMPONENT_DEFINE ("Mdrr");

namespace ns3 {

/* ... */

NS_OBJECT_ENSURE_REGISTERED (Mdrr);

TypeId
Mdrr::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::Mdrr")
    .SetParent<Object> ()
    .AddConstructor<Mdrr> ()
    .AddAttribute ("MdrrMode", "MDRR Mode: (ALTERNATE,STRICT).",
                   EnumValue (ALTERNATE),
                   MakeEnumAccessor (&Mdrr::m_MdrrMode),
                   MakeEnumChecker (ALTERNATE, "ALTERNATE",
                                    STRICT, "STRICT"))
    .AddAttribute ("MTU", "MTU of the interaface.",
                   UintegerValue (1500),
                   MakeUintegerAccessor (&Mdrr::m_Mtu),
                   MakeUintegerChecker<uint16_t> ())
    .AddAttribute ("EFWeight", "EF queue weight.",
                   UintegerValue (1),
                   MakeUintegerAccessor (&Mdrr::m_EFWeight),
                   MakeUintegerChecker<uint32_t> ())
    .AddAttribute ("AF1Weight", "AF1 queue weight.",
                   UintegerValue (1),
                   MakeUintegerAccessor (&Mdrr::m_AF1Weight),
                   MakeUintegerChecker<uint32_t> ())
    .AddAttribute ("AF2Weight", "AF2 queue weight.",
                   UintegerValue (1),
                   MakeUintegerAccessor (&Mdrr::m_AF2Weight),
                   MakeUintegerChecker<uint32_t> ())
    .AddAttribute ("AF3Weight", "AF3 queue weight.",
                   UintegerValue (1),
                   MakeUintegerAccessor (&Mdrr::m_AF3Weight),
                   MakeUintegerChecker<uint32_t> ())
    .AddAttribute ("AF4Weight", "AF4 queue weight.",
                   UintegerValue (1),
                   MakeUintegerAccessor (&Mdrr::m_AF4Weight),
                   MakeUintegerChecker<uint32_t> ())
    .AddAttribute ("BEWeight", "BE queue weight.",
                   UintegerValue (1),
                   MakeUintegerAccessor (&Mdrr::m_BEWeight),
                   MakeUintegerChecker<uint32_t> ())
    ;
  return tid;
}

// Constructor
Mdrr::Mdrr ()
{
  NS_LOG_FUNCTION_NOARGS ();

  m_Ready = false;
}

// Destructor
Mdrr::~Mdrr ()
{
  NS_LOG_FUNCTION_NOARGS ();
}

// SetDiffServ
void
Mdrr::SetDiffServ (Ptr<DiffServ> p)
{
  m_DiffServ = p->GetObject<DiffServ> ();
}

// SetWeights
void
Mdrr::SetWeights (uint32_t EF,
                  uint32_t AF1,
                  uint32_t AF2,
                  uint32_t AF3,
                  uint32_t AF4,
                  uint32_t BE)
{
  m_EFWeight = EF;
  m_AF1Weight = AF1;
  m_AF2Weight = AF2;
  m_AF3Weight = AF3;
  m_AF4Weight = AF4;
  m_BEWeight = BE;
}

// MdrrInit
//
// Formula used in quantum calculation:
// Quantum = MTU + (weight - 1) * 512
void
Mdrr::MdrrInit (void)
{
  m_EFData.Ready = false;
  m_EFData.DeficitCounter = 0;
  m_EFData.QuantumValue = m_Mtu + ((m_EFWeight - 1) * 512);

  m_AF1Data.Ready = false;
  m_AF1Data.DeficitCounter = 0;
  m_AF1Data.QuantumValue = m_Mtu + ((m_AF1Weight - 1) * 512);

  m_AF2Data.Ready = false;
  m_AF2Data.DeficitCounter = 0;
  m_AF2Data.QuantumValue = m_Mtu + ((m_AF2Weight - 1) * 512);

  m_AF3Data.Ready = false;
  m_AF3Data.DeficitCounter = 0;
  m_AF3Data.QuantumValue = m_Mtu + ((m_AF3Weight - 1) * 512);

  m_AF4Data.Ready = false;
  m_AF4Data.DeficitCounter = 0;
  m_AF4Data.QuantumValue = m_Mtu + ((m_AF4Weight - 1) * 512);

  m_BEData.Ready = false;
  m_BEData.DeficitCounter = 0;
  m_BEData.QuantumValue = m_Mtu + ((m_BEWeight - 1) * 512);

  NS_LOG_DEBUG ("EF weight: " << m_EFWeight << ", EF quantum: " << m_EFData.QuantumValue);
  NS_LOG_DEBUG ("AF1 weight: " << m_AF1Weight << ", AF1 quantum: " << m_AF1Data.QuantumValue);
  NS_LOG_DEBUG ("AF2 weight: " << m_AF2Weight << ", AF2 quantum: " << m_AF2Data.QuantumValue);
  NS_LOG_DEBUG ("AF3 weight: " << m_AF3Weight << ", AF3 quantum: " << m_AF3Data.QuantumValue);
  NS_LOG_DEBUG ("AF4 weight: " << m_AF4Weight << ", AF4 quantum: " << m_AF4Data.QuantumValue);
  NS_LOG_DEBUG ("BE weight: " << m_BEWeight << ", BE quantum: " << m_BEData.QuantumValue);

  m_AlternateCur = EF;
  m_AlternatePrev = BE;
  m_StrictCur = AF1;
}

// GetMode
Mdrr::MdrrMode
Mdrr::GetMode (void)
{
  return m_MdrrMode;
}

// SetMode
void
Mdrr::SetMode (Mdrr::MdrrMode mode)
{
  m_MdrrMode = mode;
}

// MdrrSchedule
// MDRR's main method.
// It is called in order to schedule a packet.
Ptr<Packet>
Mdrr::MdrrSchedule (void)
{
  if (!m_Ready)
    {
      MdrrInit ();
      m_Ready = true;
    }

  Ptr<Packet> p;

  switch (m_MdrrMode)
    {
      case ALTERNATE:
        p = MdrrAlternate ();
      break;
      case STRICT:
        p = MdrrStrict ();
      break;
    }

  return p;
}

// AlternateRoundRobin
// MDRR Alternate: Round robin algoritm.
void
Mdrr::AlternateRoundRobin (void)
{
  if (m_AlternateCur == EF && m_AlternatePrev == BE)
    {
      m_AlternateCur = AF1;
      m_AlternatePrev = EF;
    }
  else if (m_AlternateCur == EF && m_AlternatePrev == AF1)
    {
      m_AlternateCur = AF2;
      m_AlternatePrev = EF;
    }
  else if (m_AlternateCur == EF && m_AlternatePrev == AF2)
    {
      m_AlternateCur = AF3;
      m_AlternatePrev = EF;
    }
  else if (m_AlternateCur == EF && m_AlternatePrev == AF3)
    {
      m_AlternateCur = AF4;
      m_AlternatePrev = EF;
    }
  else if (m_AlternateCur == EF && m_AlternatePrev == AF4)
    {
      m_AlternateCur = BE;
      m_AlternatePrev = EF;
    }
  else if (m_AlternateCur == AF1)
    {
      m_AlternateCur = EF;
      m_AlternatePrev = AF1;
    }
  else if (m_AlternateCur == AF2)
    {
      m_AlternateCur = EF;
      m_AlternatePrev = AF2;
    }
  else if (m_AlternateCur == AF3)
    {
      m_AlternateCur = EF;
      m_AlternatePrev = AF3;
    }
  else if (m_AlternateCur == AF4)
    {
      m_AlternateCur = EF;
      m_AlternatePrev = AF4;
    }
  else if (m_AlternateCur == BE)
    {
      m_AlternateCur = EF;
      m_AlternatePrev = BE;
    }
}

// MdrrAlternate
// MDRR Alternate: Dequeue algorithm.
Ptr<Packet>
Mdrr::MdrrAlternate (void)
{
  NS_LOG_DEBUG (">>MDRR Alternate<<");

  Ptr<Packet> p;

  // Check if queues are empty.
  if (m_DiffServ->EmptyEF () &&
      m_DiffServ->EmptyAF1 () &&
      m_DiffServ->EmptyAF2 () &&
      m_DiffServ->EmptyAF3 () &&
      m_DiffServ->EmptyAF4 () &&
      m_DiffServ->EmptyBE ())
    {
      NS_LOG_DEBUG ("All queues are empty.");
      p = 0;
      return p;
    }

  // At least one queue has packets.
  // Loop through them until a packet is found.
  // In each loop, AlternateRoundRobin() method is called,
  // in order to find next queue in turn.
  bool done = false;
  while (!done)
    {
      if (!done && m_AlternateCur == EF)
        {
          if (!m_DiffServ->EmptyEF ())
            {
              // Initialize deficit counter for the queue.
              if (!m_EFData.Ready)
                {
                  m_EFData.DeficitCounter += m_EFData.QuantumValue;
                  m_EFData.Ready = true;
                }
              if (m_EFData.DeficitCounter > 0) // Dequeue packet.
                {
                  NS_LOG_DEBUG ("EF queue turn.");
                  p = m_DiffServ->EFDequeue ();
                  m_EFData.DeficitCounter -= p->GetSize ();
                  done = true;
                }
              else
                // Deficit counter is 0 or less.
                // Initialize deficit counter next time.
                {
                  m_EFData.Ready = false;
                }
            }
          else
            // Queue is empty.
            // Deficit counter set to 0.
            // Will be initialized at the next turn.
            {
              m_EFData.DeficitCounter = 0;
              m_EFData.Ready = false;
            }
        }
      if (!done && m_AlternateCur == AF1)
        {
          if (!m_DiffServ->EmptyAF1 ())
            {
              if (!m_AF1Data.Ready)
                {
                  m_AF1Data.DeficitCounter += m_AF1Data.QuantumValue;
                  m_AF1Data.Ready = true;
                }
              if (m_AF1Data.DeficitCounter > 0)
                {
                  NS_LOG_DEBUG ("AF1 queue turn.");
                  p = m_DiffServ->AF1Dequeue ();
                  m_AF1Data.DeficitCounter -= p->GetSize ();
                  done = true;
                }
              else
                {
                  m_AF1Data.Ready = false;
                }
            }
          else
            {
              m_AF1Data.DeficitCounter = 0;
              m_AF1Data.Ready = false;
            }
        }
      if (!done && m_AlternateCur == AF2)
        {
          if (!m_DiffServ->EmptyAF2 ())
            {
              if (!m_AF2Data.Ready)
                {
                  m_AF2Data.DeficitCounter += m_AF2Data.QuantumValue;
                  m_AF2Data.Ready = true;
                }
              if (m_AF2Data.DeficitCounter > 0)
                {
                  NS_LOG_DEBUG ("AF2 queue turn.");
                  p = m_DiffServ->AF2Dequeue ();
                  m_AF2Data.DeficitCounter -= p->GetSize ();
                  done = true;
                }
              else
                {
                   m_AF2Data.Ready = false;
                }
            }
          else
            {
              m_AF2Data.DeficitCounter = 0;
              m_AF2Data.Ready = false;
            }
        }
      if (!done && m_AlternateCur == AF3)
        {
          if (!m_DiffServ->EmptyAF3 ())
            {
              if (!m_AF3Data.Ready)
                {
                  m_AF3Data.DeficitCounter += m_AF3Data.QuantumValue;
                  m_AF3Data.Ready = true;
                }
              if (m_AF3Data.DeficitCounter > 0)
                {
                  NS_LOG_DEBUG ("AF3 queue turn.");
                  p = m_DiffServ->AF3Dequeue ();
                  m_AF3Data.DeficitCounter -= p->GetSize ();
                  done = true;
                }
              else
                {
                  m_AF3Data.Ready = false;
                }
            }
          else
            {
              m_AF3Data.DeficitCounter = 0;
              m_AF3Data.Ready = false;
            }
        }
      if (!done && m_AlternateCur == AF4)
        {
          if (!m_DiffServ->EmptyAF4 ())
            {
              if (!m_AF4Data.Ready)
                {
                  m_AF4Data.DeficitCounter += m_AF4Data.QuantumValue;
                  m_AF4Data.Ready = true;
                }
              if (m_AF4Data.DeficitCounter > 0)
                {
                  NS_LOG_DEBUG ("AF4 queue turn.");
                  p = m_DiffServ->AF4Dequeue ();
                  m_AF4Data.DeficitCounter -= p->GetSize ();
                  done = true;
                }
              else
                {
                  m_AF4Data.Ready = false;
                }
            }
          else
            {
              m_AF4Data.DeficitCounter = 0;
              m_AF4Data.Ready = false;
            }
        }
      if (!done && m_AlternateCur == BE)
        {
          if (!m_DiffServ->EmptyBE ())
            {
              if (!m_BEData.Ready)
                {
                  m_BEData.DeficitCounter += m_BEData.QuantumValue;
                  m_BEData.Ready = true;
                }
              if (m_BEData.DeficitCounter > 0)
                {
                  NS_LOG_DEBUG ("BE queue turn.");
                  p = m_DiffServ->BEDequeue ();
                  m_BEData.DeficitCounter -= p->GetSize ();
                  done = true;
                }
              else
                {
                  m_BEData.Ready = false;
                }
            }
          else
            {
              m_BEData.DeficitCounter = 0;
              m_BEData.Ready = false;
            }
        }

      AlternateRoundRobin ();
    } // while

  return p;
}

// StrictRoundRobin
// MDRR Strict: Round robin algoritm.
void
Mdrr::StrictRoundRobin (void)
{
  if (m_StrictCur == AF1)
    m_StrictCur = AF2;
  else if (m_StrictCur == AF2)
    m_StrictCur = AF3;
  else if (m_StrictCur == AF3)
    m_StrictCur = AF4;
  else if (m_StrictCur == AF4)
    m_StrictCur = BE;
  else if (m_StrictCur == BE)
    m_StrictCur = AF1;
}

// MdrrStrict
// MDRR Strict: Dequeue algorithm.
Ptr<Packet>
Mdrr::MdrrStrict (void)
{
  NS_LOG_DEBUG (">>MDRR Strict<<");

  Ptr<Packet> p;

  // Check if queues are empty.
  if (m_DiffServ->EmptyEF () &&
      m_DiffServ->EmptyAF1 () &&
      m_DiffServ->EmptyAF2 () &&
      m_DiffServ->EmptyAF3 () &&
      m_DiffServ->EmptyAF4 () &&
      m_DiffServ->EmptyBE ())
    {
      NS_LOG_DEBUG ("All queues are empty.");
      p = 0;
      return p;
    }

  // Check if high priority queue has a packet.
  if (!m_DiffServ->EmptyEF ())
    {
      NS_LOG_DEBUG ("EF queue turn.");
      p = m_DiffServ->EFDequeue ();
      m_EFData.DeficitCounter -= p->GetSize ();
      return p;
    }

  // EF queue was empty.
  // Loop through the rest of the queues until a packet is found.
  // In each loop, StrictRoundRobin() method is called,
  // in order to find the next queue in turn.
  bool done = false;
  while (!done)
    {
      if (!done && m_StrictCur == AF1)
        {
          if (!m_DiffServ->EmptyAF1 ())
            {
              if (!m_AF1Data.Ready)
                {
                  m_AF1Data.DeficitCounter += m_AF1Data.QuantumValue;
                  m_AF1Data.Ready = true;
                }
              if (m_AF1Data.DeficitCounter > 0)
                {
                  NS_LOG_DEBUG ("AF1 queue turn.");
                  p = m_DiffServ->AF1Dequeue ();
                  m_AF1Data.DeficitCounter -= p->GetSize ();
                  done = true;
                }
              else
                {
                  m_AF1Data.Ready = false;
                }
            }
          else
            {
              m_AF1Data.DeficitCounter = 0;
              m_AF1Data.Ready = false;
            }
        }
      if (!done && m_StrictCur == AF2)
        {
          if (!m_DiffServ->EmptyAF2 ())
            {
              if (!m_AF2Data.Ready)
                {
                  m_AF2Data.DeficitCounter += m_AF2Data.QuantumValue;
                  m_AF2Data.Ready = true;
                }
              if (m_AF2Data.DeficitCounter > 0)
                {
                  NS_LOG_DEBUG ("AF2 queue turn.");
                  p = m_DiffServ->AF2Dequeue ();
                  m_AF2Data.DeficitCounter -= p->GetSize ();
                  done = true;
                }
              else
                {
                   m_AF2Data.Ready = false;
                }
            }
          else
            {
              m_AF2Data.DeficitCounter = 0;
              m_AF2Data.Ready = false;
            }
        }
      if (!done && m_StrictCur == AF3)
        {
          if (!m_DiffServ->EmptyAF3 ())
            {
              if (!m_AF3Data.Ready)
                {
                  m_AF3Data.DeficitCounter += m_AF3Data.QuantumValue;
                  m_AF3Data.Ready = true;
                }
              if (m_AF3Data.DeficitCounter > 0)
                {
                  NS_LOG_DEBUG ("AF2 queue turn.");
                  p = m_DiffServ->AF3Dequeue ();
                  m_AF3Data.DeficitCounter -= p->GetSize ();
                  done = true;
                }
              else
                {
                  m_AF3Data.Ready = false;
                }
            }
          else
            {
              m_AF3Data.DeficitCounter = 0;
              m_AF3Data.Ready = false;
            }
        }
      if (!done && m_StrictCur == AF4)
        {
          if (!m_DiffServ->EmptyAF4 ())
            {
              if (!m_AF4Data.Ready)
                {
                  m_AF4Data.DeficitCounter += m_AF4Data.QuantumValue;
                  m_AF4Data.Ready = true;
                }
              if (m_AF4Data.DeficitCounter > 0)
                {
                  NS_LOG_DEBUG ("AF4 queue turn.");
                  p = m_DiffServ->AF4Dequeue ();
                  m_AF4Data.DeficitCounter -= p->GetSize ();
                  done = true;
                }
              else
                {
                  m_AF4Data.Ready = false;
                }
            }
          else
            {
              m_AF4Data.DeficitCounter = 0;
              m_AF4Data.Ready = false;
            }
        }
      if (!done && m_StrictCur == BE)
        {
          if (!m_DiffServ->EmptyBE ())
            {
              if (!m_BEData.Ready)
                {
                  m_BEData.DeficitCounter += m_BEData.QuantumValue;
                  m_BEData.Ready = true;
                }
              if (m_BEData.DeficitCounter > 0)
                {
                  NS_LOG_DEBUG ("BE queue turn.");
                  p = m_DiffServ->BEDequeue ();
                  m_BEData.DeficitCounter -= p->GetSize ();
                  done = true;
                }
              else
                {
                  m_BEData.Ready = false;
                }
            }
          else
            {
              m_BEData.DeficitCounter = 0;
              m_BEData.Ready = false;
            }
        }

      StrictRoundRobin ();
    } // while

  return p;
}

} //namespace ns3
