/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2013 University of Patras,
 * Department of Computer Engineering & Informatics
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * Author: Christos Nianiakas <cnianiak_at_gmail_dot_com>
 */

#ifndef __MDRR_H__
#define __MDRR_H__

#include "diffserv.h"
#include "ns3/object.h"
#include "ns3/packet.h"

namespace ns3 {

/**
 * \brief Modified Deficit Round Robin (MDRR) algorithm implementation
 *
 * Used by the implemented DiffServ architecture
 */

class Mdrr : public Object
{
public:
  static TypeId GetTypeId (void);

/**
 * \brief Mdrr Constructor
 *
 * Creates an Mdrr instance into ALTERNATE mode
 */
  Mdrr ();
  virtual ~Mdrr ();

  // Mdrr operational modes
  enum MdrrMode
    {
      ALTERNATE,
      STRICT
    };

  // Mdrr data used to service the queues
  struct MdrrData
    {
      bool Ready;
      uint32_t QuantumValue;
      uint32_t DeficitCounter;
    };

/**
 * Set DiffServ class instance that holds the queues for Mdrr to manage
 *
 * \param p Smart pointer to DiffServ class instance
 */
  void SetDiffServ (Ptr<DiffServ> p);
/**
 * Set the weight of each PHB queue
 *
 * \param EF Expedited Forwarding PHB queue
 * \param AF1 Assured Forwarding class 1 PHB queue
 * \param AF2 Assured Forwarding class 2 PHB queue
 * \param AF3 Assured Forwarding class 3 PHB queue
 * \param AF4 Assured Forwarding class 4 PHB queue
 * \param BE Best Effort PHB queue
 */
  void SetWeights (uint32_t EF,
                   uint32_t AF1,
                   uint32_t AF2,
                   uint32_t AF3,
                   uint32_t AF4,
                   uint32_t BE);
/**
 * Initialize Mdrr
 */
  void MdrrInit (void);
/**
 * Get the Mdrr operational mode
 *
 * \return Returns Mdrr mode enum type
 */
  Mdrr::MdrrMode GetMode (void);
/**
 * Set the Mdrr operational mode
 *
 * \param mode Mdrr mode to set
 */
  void SetMode (Mdrr::MdrrMode mode);
/**
 * Main schedule method for both operational modes
 *
 * \return Returns a Packet smart pointer
 */
  Ptr<Packet> MdrrSchedule (void);
/**
 * Finds the next queue in turn for ALTERNATE mode
 */
  void AlternateRoundRobin (void);
/**
 * Dequeue method for ALTERNATE mode
 *
 * \return Returns a Packet smart pointer
 */
  Ptr<Packet> MdrrAlternate (void);
/**
 * Finds the next queue in turn for STRICT mode
 */
  void StrictRoundRobin (void);
/**
 * Dequeue method for STRICT mode
 *
 * \return Returns a Packet smart pointer
 */
  Ptr<Packet> MdrrStrict (void);

private:
  bool m_Ready; // Flag to initialize Mdrr once
  uint16_t m_Mtu; // MTU of the interface
  MdrrMode m_MdrrMode; // Mdrr operational mode
  Ptr<DiffServ> m_DiffServ; // Smart pointer to DiffServ class instance

  // Service info for available queues
  MdrrData m_EFData;
  uint32_t m_EFWeight;

  MdrrData m_AF1Data;
  uint32_t m_AF1Weight;

  MdrrData m_AF2Data;
  uint32_t m_AF2Weight;

  MdrrData m_AF3Data;
  uint32_t m_AF3Weight;

  MdrrData m_AF4Data;
  uint32_t m_AF4Weight;

  MdrrData m_BEData;
  uint32_t m_BEWeight;

  // Track of service order
  DsQueues m_AlternateCur;
  DsQueues m_AlternatePrev;
  DsQueues m_StrictCur;
};

} //namespace ns3

#endif /* __MDRR_H__ */
