/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2013 University of Patras,
 * Department of Computer Engineering & Informatics
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * Author: Christos Nianiakas <cnianiak_at_gmail_dot_com>
 */

#include "policy.h"
#include "ns3/log.h"
#include "ns3/ppp-header.h"
#include "ns3/random-variable.h"

NS_LOG_COMPONENT_DEFINE ("Policy");

namespace ns3 {

/* ... */

NS_OBJECT_ENSURE_REGISTERED (Policy);

TypeId
Policy::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::Policy")
    .SetParent<Object> ()
    .AddConstructor<Policy> ()
    ;
  return tid;
}

// Constructor
Policy::Policy ()
{
  NS_LOG_FUNCTION_NOARGS ();

  // Initialize some of the metrics used by the algorithms.
  m_Metrics.arrivalTime = 0;
  m_Metrics.cBucket = 0;
  m_Metrics.eBucket = 0;
  m_Metrics.pBucket = 0;
  m_Metrics.winLen = 1;
}

// Destructor
Policy::~Policy ()
{
  NS_LOG_FUNCTION_NOARGS ();
}

// SetPolicy
// Set the policy actions and metrics.
void
Policy::SetPolicy (Policy::Actions actions, Policy::Metrics metrics)
{
  m_Actions = actions;
  m_Metrics = metrics;
}

// GetActions
Policy::Actions
Policy::GetActions (void)
{
  return m_Actions;
}

// PacketSize
// Calculate packet size,
// without considering data link layer header.
uint32_t
Policy::PacketSize (Ptr<Packet> p)
{
  Ptr<Packet> q = p->Copy ();
  uint32_t bytesRead = 0;
  PppHeader pppHeader;
  bytesRead = q->RemoveHeader (pppHeader);
  if (bytesRead > 0)
    {
      NS_LOG_DEBUG ("PPP Header found.");
    }
  else
    {
      NS_LOG_DEBUG ("Failed to read PPP header.");
      return Red;
    }

  return q->GetSize ();
}

// Meter
// Main method.
// Called in order to apply a policer for a packet.
Policy::Conformance
Policy::Meter (Ptr<Packet> p)
{
  NS_LOG_DEBUG (">> Meter <<");

  if (m_Actions.Type == TSW2CM)
    {
      NS_LOG_DEBUG ("TSW2CM policer.");
      return TSW2CMMeter (p);
    }
  else if (m_Actions.Type == TSW3CM)
    {
      NS_LOG_DEBUG ("TSW3CM policer.");
      return TSW3CMMeter (p);
    }
  else if (m_Actions.Type == TokenBucket)
    {
      NS_LOG_DEBUG ("TokenBucket policer.");
      return TokenBucketMeter (p);
    }
  else if (m_Actions.Type == srTCM)
    {
      NS_LOG_DEBUG ("srTCM policer.");
      return SrTcmMeter (p);
    }
  else if (m_Actions.Type == trTCM)
    {
      NS_LOG_DEBUG ("trTCM policer.");
      return TrTcmMeter (p);
    }

  // Null policy
  NS_LOG_DEBUG ("Null policer.");
  return Green;
}

/**
 * Algorithms ported from ns-2.
 * Found in: ns-2.35/diffserv/dsPolicy.{h,cc}
 *
 * For each algorithm:
 * both applyPolicer() and applyMeter() methods
 * fussed in a single "Meter" method.
 */

// TSW2CMMeter
Policy::Conformance
Policy::TSW2CMMeter (Ptr<Packet> p)
{
  double bytesInTSW = m_Metrics.avgRate * m_Metrics.winLen;
  uint32_t PktSize = PacketSize (p);
  double newBytes = bytesInTSW + PktSize;
  double now = Simulator::Now ().GetSeconds ();

  m_Metrics.avgRate = newBytes / (now - m_Metrics.arrivalTime + m_Metrics.winLen);
  m_Metrics.arrivalTime = now;

  UniformVariable x (0.0, 1.0);
  if ((m_Metrics.avgRate > m_Metrics.CIR)
      && (x.GetValue () <= (1-(m_Metrics.CIR/m_Metrics.avgRate))))
    {
      return Red;
    }
  else
    return Green;
}

// TSW3CMMeter
Policy::Conformance
Policy::TSW3CMMeter (Ptr<Packet> p)
{
  double bytesInTSW = m_Metrics.avgRate * m_Metrics.winLen;
  uint32_t PktSize = PacketSize (p);
  double newBytes = bytesInTSW + PktSize;
  double now = Simulator::Now ().GetSeconds ();

  m_Metrics.avgRate = newBytes / (now - m_Metrics.arrivalTime + m_Metrics.winLen);
  m_Metrics.arrivalTime = now;

  UniformVariable x (0.0, 1.0);
  double rand = m_Metrics.avgRate * (1.0 - x.GetValue ());

  if (rand > m_Metrics.PIR)
    return Red;
  else if (rand > m_Metrics.CIR)
    return Yellow;
  else
    return Green;
}
// TokenBucketMeter
Policy::Conformance
Policy::TokenBucketMeter (Ptr<Packet> p)
{
  double now = Simulator::Now ().GetSeconds ();
  double TokenBytes = m_Metrics.CIR * (now - m_Metrics.arrivalTime);

  if (m_Metrics.cBucket + TokenBytes <= m_Metrics.CBS)
    m_Metrics.cBucket += TokenBytes;
  else
    m_Metrics.cBucket = m_Metrics.CBS;

  m_Metrics.arrivalTime = now;

  uint32_t PktSize = PacketSize (p);

  if ((m_Metrics.cBucket - PktSize) >= 0)
    {
      m_Metrics.cBucket -= PktSize;
      return Green;
    }
  else
      return Red;
}
// SrTcmMeter
Policy::Conformance
Policy::SrTcmMeter (Ptr<Packet> p)
{
  double now = Simulator::Now ().GetSeconds ();
  double TokenBytes = m_Metrics.CIR * (now - m_Metrics.arrivalTime);

  if (m_Metrics.cBucket + TokenBytes <= m_Metrics.CBS)
    m_Metrics.cBucket += TokenBytes;
  else
    {
      TokenBytes = TokenBytes - (m_Metrics.CBS - m_Metrics.cBucket);

      m_Metrics.cBucket = m_Metrics.CBS;
      if (m_Metrics.eBucket + TokenBytes <= m_Metrics.EBS)
        m_Metrics.eBucket += TokenBytes;
      else
        m_Metrics.eBucket = m_Metrics.EBS;
    }

  m_Metrics.arrivalTime = now;

  uint32_t PktSize = PacketSize (p);

  if ((m_Metrics.cBucket - PktSize) >= 0)
    {
      m_Metrics.cBucket -= PktSize;
      return Green;
    }
  else
    {
      if ((m_Metrics.eBucket - PktSize) >= 0)
        {
          m_Metrics.eBucket -= PktSize;
          return Yellow;
        }
      else
        return Red;
    }
}

// TrTcmMeter
Policy::Conformance
Policy::TrTcmMeter (Ptr<Packet> p)
{
  double now = Simulator::Now ().GetSeconds ();
  double TokenBytes = m_Metrics.CIR * (now - m_Metrics.arrivalTime);

  if (m_Metrics.cBucket + TokenBytes <= m_Metrics.CBS)
    m_Metrics.cBucket += TokenBytes;
  else
    m_Metrics.cBucket = m_Metrics.CBS;

  TokenBytes = m_Metrics.PIR * (now - m_Metrics.arrivalTime);
  if (m_Metrics.pBucket + TokenBytes <= m_Metrics.PBS)
    m_Metrics.pBucket += TokenBytes;
  else
    m_Metrics.pBucket = m_Metrics.PBS;

  m_Metrics.arrivalTime = now;

  uint32_t PktSize = PacketSize (p);

  if ((m_Metrics.pBucket - PktSize) < 0)
    return Red;
  else
    {
      if ((m_Metrics.cBucket - PktSize) < 0)
        {
          m_Metrics.pBucket -= PktSize;
          return Yellow;
        }
      else
        {
          m_Metrics.cBucket -= PktSize;
          m_Metrics.pBucket -= PktSize;
          return Green;
        }
    }
}

} //namespace ns3
