/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2013 University of Patras,
 * Department of Computer Engineering & Informatics
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * Author: Christos Nianiakas <cnianiak_at_gmail_dot_com>
 */

#ifndef __POLICY_H__
#define __POLICY_H__

#include "ns3/object.h"
#include "ns3/ipv4-header.h"
#include "ns3/packet.h"
#include "ns3/simulator.h"

namespace ns3 {

/**
 * \brief Policy mechanisms (ported from ns-2)
 *
 * Used by the implemented DiffServ architecture
 */


class Policy : public Object
{
public:
  static TypeId GetTypeId (void);

/**
 * \brief Policy Constructor
 */
  Policy ();
  virtual ~Policy ();

  // Levels of conformance
  enum Conformance
    {
      Green,
      Yellow,
      Red
    };

  // Available policy mechanisms
  enum Type
    {
      Null,
      TSW2CM,
      TSW3CM,
      TokenBucket,
      srTCM,
      trTCM
    };

  // Metrics used by the implemented policy mechanisms
  struct Metrics
    {
      double CIR; // Committed Information Rate
      double CBS; // Committed Burst Size
      double cBucket; // Committed Bucket
      double EBS; // Excess Burst Size
      double eBucket; // Excess Bucket
      double PIR; // Peak Information Rate
      double PBS; // Peak Burst Size
      double pBucket; // Peak Bucket
      double avgRate; // Average Rate
      double winLen; // Windows Length
      double arrivalTime; // Packet Arrival time
    };

/**
 * Available actions.
 * Drop a packet or assign a DSCP value to it
 */
  enum Mode
    {
      Drop,
      Dscp
    };
/**
 * Set the action to take
 * and the DSCP value for it (if necessary)
 */
  struct Action
    {
      Policy::Mode Mode;
      Ipv4Header::DscpType Dscp;
    };
/**
 * Set the type of policer to use
 * and (up to) three conformance actions
 */
  struct Actions
    {
      Policy::Type Type;
      Policy::Action Green;
      Policy::Action Yellow;
      Policy::Action Red;
    };

/**
 * Setup the policy to be used
 *
 * \param actions The acitons to take for each conformance level
 * \param metrics The metrics to be used by the selected mechanism
 */
  void SetPolicy (Policy::Actions actions, Policy::Metrics metrics);
/**
 * Get the actions in use
 *
 * \return Returns a Policy::Actions struct
 */
  Policy::Actions GetActions (void);
/**
 * Calculates a packet's size (without considering the ppp-Header)
 *
 * \param p Packet smart pointer to calculate
 * \return Returns packet's size
 */
  uint32_t PacketSize (Ptr<Packet> p);
/**
 * Entry method for the implemented mechanisms
 *
 * \param p Packet smart pointer to meter and police
 * \return Returns packet's conformance level
 */
  Policy::Conformance Meter (Ptr<Packet> p);
/**
 * Implemented mechanisms
 *
 * \param p Packet smart pointer to meter and police
 * \return Returns packet's conformance level
 */
  Policy::Conformance TSW2CMMeter (Ptr<Packet> p);
  Policy::Conformance TSW3CMMeter (Ptr<Packet> p);
  Policy::Conformance TokenBucketMeter (Ptr<Packet> p);
  Policy::Conformance SrTcmMeter (Ptr<Packet> p);
  Policy::Conformance TrTcmMeter (Ptr<Packet> p);

private:
  Policy::Actions m_Actions;
  Policy::Metrics m_Metrics;
};

} //namespace ns3

#endif /* __POLICY_H__ */
