/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2013 University of Patras,
 * Department of Computer Engineering & Informatics
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * Author: Christos Nianiakas <cnianiak_at_gmail_dot_com>
 */

#include "traffic-helper.h"
#include "ns3/log.h"
#include "ns3/string.h"
#include "ns3/uinteger.h"
#include "ns3/inet-socket-address.h"
#include "ns3/on-off-helper.h"
#include "ns3/udp-client-server-helper.h"

namespace ns3 {

/* ... */

NS_LOG_COMPONENT_DEFINE ("TrafficHelper");

// VoIpHelper
NS_OBJECT_ENSURE_REGISTERED (VoIpHelper);

TypeId
VoIpHelper::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::VoIpHelper")
    .SetParent<Object> ()
    .AddConstructor<VoIpHelper> ()
    .AddAttribute ("Rate", "Data rate of application.",
                   DataRateValue (DataRate ("80Kbps")),
                   MakeDataRateAccessor (&VoIpHelper::m_Rate),
                   MakeDataRateChecker ())
    .AddAttribute ("PacketSize", "Packet size of the application.",
                   UintegerValue (160),
                   MakeUintegerAccessor (&VoIpHelper::m_PacketSize),
                   MakeUintegerChecker<uint32_t> ())
    ;
  return tid;
}

VoIpHelper::VoIpHelper ()
{
  m_Duration = false;
  m_Destination = false;
}

VoIpHelper::~VoIpHelper ()
{
}

void
VoIpHelper::SetDuration (double start, double stop)
{
  m_Start = start;
  m_Stop = stop;
  m_Duration = true;
}

void
VoIpHelper::SetDestination (Ipv4Address address, uint16_t port)
{
  m_Address = address;
  m_Port = port;
  m_Destination = true;
}

void
VoIpHelper::Install (Ptr<Node> n)
{
  NS_ASSERT (m_Duration == true && m_Destination == true);
  OnOffHelper onoff ("ns3::UdpSocketFactory", InetSocketAddress (m_Address, m_Port));
  onoff.SetAttribute ("DataRate", DataRateValue (m_Rate));
  onoff.SetAttribute ("PacketSize", UintegerValue (m_PacketSize));
  onoff.SetAttribute ("OnTime", StringValue ("ns3::ExponentialRandomVariable[Mean=1.448]"));
  onoff.SetAttribute ("OffTime", StringValue ("ns3::ExponentialRandomVariable[Mean=2.29]"));
  ApplicationContainer app = onoff.Install (n);
  app.Start (Seconds (m_Start));
  app.Stop (Seconds (m_Stop));
}

// VideoTraceHelper
NS_OBJECT_ENSURE_REGISTERED (VideoTraceHelper);

TypeId
VideoTraceHelper::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::VideoTraceHelper")
    .SetParent<Object> ()
    .AddConstructor<VideoTraceHelper> ()
    .AddAttribute ("MaxPacketSize", "Set max packet size.",
                   UintegerValue (200),
                   MakeUintegerAccessor (&VideoTraceHelper::m_MaxPacketSize),
                   MakeUintegerChecker<uint32_t> ())
    ;
  return tid;
}

VideoTraceHelper::VideoTraceHelper ()
{
  m_Duration = false;
  m_Destination = false;
  m_Trace = false;
}

VideoTraceHelper::~VideoTraceHelper ()
{
}

void
VideoTraceHelper::SetDuration (double start, double stop)
{
  m_Start = start;
  m_Stop = stop;
  m_Duration = true;
}

void
VideoTraceHelper::SetDestination (Ipv4Address address, uint16_t port)
{
  m_Address = address;
  m_Port = port;
  m_Destination = true;
}

void
VideoTraceHelper::SetTrace (std::string source)
{
  m_TraceSource = source;
  m_Trace = true;
}

void
VideoTraceHelper::Install (Ptr<Node> n)
{
  NS_ASSERT (m_Duration == true && m_Destination == true && m_Trace == true);
  UdpTraceClientHelper udptrace (m_Address, m_Port, m_TraceSource);
  udptrace.SetAttribute ("MaxPacketSize", UintegerValue (m_MaxPacketSize));
  ApplicationContainer app = udptrace.Install (n);
  app.Start (Seconds (m_Start));
  app.Stop (Seconds (m_Stop));
}

} //namespace ns3
